//=======================================================================
// matrice.cc : une classe pour modliser les matrices 2x2 
// Copyright (C) 2004-2007 by Michael Eisermann, all rights reserved
// Institut Fourier, Universit Joseph Fourier, Grenoble, France
//
// Description :
// Ce programme donne un exemple de modlisation oriente objet,
// en implmentant les matrices 2x2 avec leurs oprations usuelles.
//=======================================================================

#include <iostream>
typedef long long int Integer;
using namespace std;

//-----------------------------------------------------------------------
// Une matrice 2x2 est dfinie par ses quatre coefficients a,b,c,d :
//   ( a b )
//   ( c d )
// On prvoit trois constructeurs : construire un multiple de l'identit
// (conversion de Integer en Matrice), ou une matrice diagonale ( partir
// de deux Integer), ou bien une matrice quelconque (quatre coeffs).
// On ajoute aussi quelques mthodes pour rdfinir les coefficients.
//-----------------------------------------------------------------------

class Matrice
{
public:
  // Les quatre coefficients d'une matrice
  Integer a,b,c,d;
  
public:
  // Construire une matrice quelconque
  Matrice( const Integer& a0, const Integer& b0, 
	   const Integer& c0, const Integer& d0 )
    : a(a0), b(b0), c(c0), d(d0) {};
  
  // Construire une matrice diagonale 
  Matrice( const Integer& s, const Integer& t )
    : a(s), b(0), c(0), d(t) {};
  
  // Construire un multiple de l'identit 
  Matrice( const Integer& t= Integer(0) )
    : a(t), b(0), c(0), d(t) {};
  
  // Remettre tous les coefficients  zro
  void clear() { a=0; b=0; c=0; d=0; };
  
  // Redfinir tous les coefficients
  void def(const Integer& a0, const Integer& b0, 
	   const Integer& c0, const Integer& d0 )
  { a=a0; b=b0; c=c0; d=d0; };
};

//-----------------------------------------------------------------------
// Oprations sur les matrices 2x2
//-----------------------------------------------------------------------

// Oprations scalaires
Matrice operator* ( const Integer& t, const Matrice& y )
{ return Matrice( t*y.a, t*y.b, t*y.c, t*y.d ); }

Matrice operator* ( const Matrice& y, const Integer& t )
{ return Matrice( y.a*t, y.b*t, y.c*t, y.d*t ); }

Matrice operator/ ( const Matrice& y, const Integer& t )
{ return Matrice( y.a/t, y.b/t, y.c/t, y.d/t ); }

Matrice operator% ( const Matrice& y, const Integer& t )
{ return Matrice( y.a%t, y.b%t, y.c%t, y.d%t ); }

// Addition, soustraction, multiplication de deux matrices
Matrice operator+ ( const Matrice& x, const Matrice& y )
{ return Matrice( x.a+y.a, x.b+y.b, x.c+y.c, x.d+y.d ); }

Matrice operator- ( const Matrice& x, const Matrice& y )
{ return Matrice( x.a-y.a, x.b-y.b, x.c-y.c, x.d-y.d ); }

Matrice operator* ( const Matrice& x, const Matrice& y )
{ return Matrice( x.a*y.a + x.b*y.c , x.a*y.b + x.b*y.d,
		  x.c*y.a + x.d*y.c , x.c*y.b + x.d*y.d ); }

// Transpose
Matrice transposee( const Matrice& x )
{ return Matrice( x.a, x.c, x.b, x.d ); }

// Dterminant
Integer det( const Matrice& x )
{ return x.a*x.d - x.b*x.c; }

// Tester l'inversibilit
bool inversible( const Matrice& x )
{
  Integer d= det(x);
  return ( d==Integer(1) || d==Integer(-1) );
}

// Tester l'inversibilit et calculer l'inverse  
bool inversible( const Matrice& x, Matrice& y )
{
  Integer d= det(x);
  if ( d!=Integer(1) && d!=Integer(-1) ) return false;
  y.def( x.d/d, -x.b/d, -x.c/d, x.a/d );
  return true;
}
  
// Division / inversion d'une matrice
Matrice operator / ( const Matrice& x, const Matrice& y )
{
  Matrice y_inverse;
  if ( !inversible( y, y_inverse ) )
    {
      cerr << "division par une matrice non inversible !" << endl;
      exit(1);
    };
  return x * y_inverse;
}

// Oprateur de sortie pour une matrice 2x2
ostream& operator << ( ostream& out, const Matrice& x )
{
  return out << "( " << x.a << ", " << x.b << "; " << x.c << ", " << x.d << ") ";
}

// Une fonction d'entre auxiliaire
bool next( istream& in, char symbole )
{
  if ( !in.good() ) return false;
  char entree;
  in >> entree;
  if ( in.fail() ) return false;
  if ( entree == symbole ) return true;
  in.putback( symbole );
  in.setstate( ios::failbit );
  return false;
}      

// Oprateur d'entre pour une matrice 2x2
istream& operator >> ( istream& in, Matrice& x )
{
  in.clear();
  x.clear();
  if ( !next( in, '(' ) ) return in;
  in >> x.a;
  if ( !next( in, ',' ) ) return in;
  in >> x.b;
  if ( !next( in, ';' ) ) return in;
  in >> x.c;
  if ( !next( in, ',' ) ) return in;
  in >> x.d;
  if ( !next( in, ')' ) ) return in;
  return in;
}

//-----------------------------------------------------------------------
// Exemple d'utilisation de la classe matrice
//-----------------------------------------------------------------------
int main()
{
  Matrice a(1,2,3,4), b;
  cout << "Bonjour et bienvenue aux matrices !\n"
       << "Ce programme permet de calculer avec les maxtrices de taille 2x2\n"
       << " coefficients entiers. Exemple de notation : " << a << endl
       << "Entrez une matrice 2x2 svp : ";
  cin >> a;
  cout << "Vous avez entre a = " << a << endl;
  cout << "Le dterminant vaut " << det(a) << endl;
  if ( inversible(a,b) )
    cout << "La matrice a est inversible," << endl
	 << "avec inverse a^-1 = " << b << "," << endl
	 << "car " << a << " * " << b << " = " << a*b << endl;
  cout << "\nVoici les puissance de a :" << endl;
  Matrice p(1);
  for ( int k=1; k<=10; ++k )
    {
      p= p*a;
      cout << "a^" << k << " = " << p << endl;
    };
  cout << "Au revoir." << endl;
}

