//=======================================================================
// karatsuba.cc : multiplication rapide selon Karatsuba
// Copyright (C) 2004-2007 Michael Eisermann, all rights reserved
// Institut Fourier, Universit Joseph Fourier, Grenoble, France
//
//  compiler avec l'option -O3 pour optimiser la rapidit. 
//
// Description :
// Soient a et b deux nombres naturels de longueur m dans une base donne.
// La mthode scolaire de la multiplication ncessite m^2 itrations pour
// calculer le produit a*b. L'algorithme de Karatsuba calcule le mme
// produit avec une complexit sous-quadratique,  savoir O(m^e) avec
// un exposant e = log(3)/log(2) = 1.585..., ce qui est nettement mieux 
// que l'exposant 2 classique.
//
// Structure de donnes :
// On utilise le type Naturel implment dans le fichier "naturel.cc",
// qui reprsente un nombre naturel par une suite de chiffres dans
// la base choisie (typiquement la base 10).
//
// Esquisse de l'algorithme :
// L'ide est simple : d'abord on dcompose les facteurs a et b comme
//   a = a0 + a1*base^n  
//   b = b0 + b1*base^n
// Ici a0 et b0 sont compris dans l'intervalle [ 0, base^n [.
// Bien videmment, le produit cherch est gal 
//   a * b = (a0*b0) + (a0*b1+a1*b0) base^n + (a1*b1) base^(2n)
// ce qui semble ncessiter quatre multiplications au total.
//
// L'astuce de Karatsuba :
// L'observation de Karatsuba est que l'on peut arriver au mme
// rsultat avec trois multiplications seulement !  On calcule
//   s = a0 * b0
//   t = a1 * b1 
//   u = (a0+a1) * (b0+b1 )  -  s  -  t
// On constate que  u = a0*b1 + a1*b0 , donc le produit cherch est
//   s  +  u base^n  +  t base^(2n)
// Les multiplications par base^n et base^(2n) ne sont que des dcalages 
// d'indices. Au total on n'utilise donc que trois multiplications.
//
// Complexit :
// La mthode de Karatsuba est facile  implmenter (voir plus bas). 
// En comparaison avec la mthode scolaire, celle-ci est plus rapide 
// pour les petits nombres, mais  long terme c'est Karatsuba qui gagne:
// chaque fois que la longueur double, le temps d'excution est multipli 
// par 4 pour la mthode scolaire, mais seulement par 3 pour Karatsuba!
// Ceci se reflte dans la complexit asymptotique: c'est O(m^2) pour 
// la mthode scolaire mais seulement O(m^1.585) pour Karatsuba.
//=======================================================================

#include "naturel.cc"          // notre implmentation faite maison
using namespace std;           // accs direct aux fonctions standard

//-----------------------------------------------------------------------
// Multiplication selon Karatsuba
// Entre : deux nombres naturels a et b
// Sortie : leur produit a*b 
//-----------------------------------------------------------------------
bool karatsuba( const Naturel& a, const Naturel& b, Naturel& produit )
{
  // Dlguer les petites multiplications  la mthode scolaire
  if ( a.size() < 40 || b.size() < 40 ) 
    {
      multiplication( a, b, produit );
      return true;
    };

  // Dterminer une taille de coupure convenable
  Indice m= max( a.size(), b.size() );
  if ( m%2 == 1 ) ++m;
  Indice n= m/2;

  // Dcouper a et b en deux moitis
  Naturel a0, a1, b0, b1, s, t, u;
  decouper( a, n, a1, a0 );
  decouper( b, n, b1, b0 );

  // Multiplication rcursive selon Karatsuba
  karatsuba( a0, b0, s );
  karatsuba( a1, b1, t );
  karatsuba( a0+a1, b0+b1, u );
  u-= s;
  u-= t;
  
  // Mettre s et t bout  bout dans produit
  produit.chiffres.clear();
  produit.chiffres.resize( a.size()+b.size(), Chiffre(0) );
  for( Indice i=0; i<s.size(); ++i ) produit.chiffres[i]  = s.chiffres[i];
  for( Indice i=0; i<t.size(); ++i ) produit.chiffres[m+i]= t.chiffres[i];
  
  // Ajouter u au milieu du produit
  Indice i= n; Chiffre retenue= 0;
  for( Indice j=0; j<u.size(); ++j, ++i )
    {
      produit.chiffres[i]+= u.chiffres[j] + retenue;
      if ( produit.chiffres[i] < base ) { retenue= 0; }
      else { produit.chiffres[i]-= base; retenue = 1; };
    };

  // Stocker une ventuelle retenue  la fin, puis raccourcir le rsultat
  if ( retenue > 0 )
    {
      while ( produit.chiffres[i] == Chiffre(base-1) )
	produit.chiffres[i++]= 0;
      ++produit.chiffres[i];
    };
  produit.raccourcir();
  return true;
}

//-----------------------------------------------------------------------
// Dcouper  n = n0 + n1 * base^e  avec  0 <= n0 < base^e  et  0 <= n1
// Entre    : un nombre naturel n et un exposant e>0
// Sortie    : la partie basse n0 et la partie haute n1 de n
// Il s'agit d'une division euclidienne par base^e.   noter que n est 
// dj reprsent dans la base souhaite, il suffit donc de recopier 
// les chiffres des parties basse et haute.
//-----------------------------------------------------------------------
void decouper( const Naturel& n, Indice e, Naturel& n1, Naturel& n0 )
{
  // Effacer les anciennes valeurs de n0 et n1
  n0.clear();
  n1.clear();

  // Cas exceptionnel: tout rentre dans la partie basse
  Indice taille= n.size();
  if ( e >= taille ) { n0= n; return; };
  
  // Copier la partie haute (comme n est normalis, n1 l'est aussi)
  n1.chiffres.resize( taille-e );
  for( Indice i=0, j=e; j<taille; ++i, ++j ) n1.chiffres[i]= n.chiffres[j];
  
  // Dterminer la partie basse en supprimant d'ventuels zros terminaux
  Indice i= e-1;
  while( i>=0 && n.chiffres[i]==Chiffre(0) ) --i;
  n0.chiffres.resize(i+1);
  for( ; i>=0; --i ) n0.chiffres[i]= n.chiffres[i];
}

