//=======================================================================
// gmp-chrono.cc : comparer la performance de Naturel avec mpz_class
// Copyright (C) 2004-2007 by Michael Eisermann, all rights reserved
// Institut Fourier, Universit Joseph Fourier, Grenoble, France
//
// Description :
// Ce petit programme compare la performance du type Naturel, fait maison,
// avec celle de la classe mpz_class, qui elle est hautement optimise.
//
// Compilation :
//  compiler avec l'option -lgmpxx pour inclure la bibliothque GMP.
// Ajouter l'option -O3 pour une compilation optimisant la rapidit. 
//
// Optimisation :
// En passant on peut tester la performance de notre compilateur 
// en matire d'optimisation. Pour ceci compiler une fois avec g++
// (sans optimisation), puis avec g++ -O1, g++ -O2, et g++ -O3 
// (optimisation de plus en plus sophistique). Sachez toutefois
// que mme le meilleur compilateur ne remplacera jamais le 
// dveloppement d'un algorithme efficace.
//=======================================================================

#include <iostream>         // dclarer l'entre-sortie standard
#include <iomanip>          // manipulateurs pour formater les flots
#include <sstream>          // conversion entre string et flot
using namespace std;        // accs direct aux fonctions standard

// #include "naturel.cc"    // dfinir le type Naturel, fait maison
#include "karatsuba.cc"     // dfinir la multiplication selon Karatsuba
#include "timer.hh"         // horloge pour chronomtrer l'excution
#include <gmpxx.h>          // accder  la bibliothque GMP pour C++
typedef mpz_class Integer;  // synonyme plus parlant pour mpz_class

#define __verification__    // option pour ajouter des vrifications

//-----------------------------------------------------------------------
// Fonctions auxiliaires
//-----------------------------------------------------------------------

Integer random( unsigned long bits )
{ 
  // Engendrer un nombre alatoire (de taille donne en bits)
  static gmp_randclass generateur( gmp_randinit_default );
  return generateur.get_z_bits( bits );
}

void affecter( Naturel& a, Integer b )
{
  // Convertir une valeur de type Integer en une suite de dcimales
  // par une division euclidienne itre.
  /*
  vector<Chiffre> suite;
  for( ; b>0; b/= base ) 
    suite.push_back( Chiffre( Integer( b % base ).get_si() ) );
  a= Naturel( suite );
  */

  // Astuce: pour des raisons d'efficacit on utilise l'affichage fourni 
  // par la bibliothque GMP, en passant par une chane de caractre.  
  ostringstream oss;
  oss << b;
  istringstream iss( oss.str() );
  iss >> a;
}

//-----------------------------------------------------------------------
// Un petit programme pour chronomtrer l'addition et la multiplication
//-----------------------------------------------------------------------

int main()
{
  cout << "\nBonjour et bienvenue aux nombres naturels !\n" 
       << "Ce petit logiciel compare la performance de mpz_class (GMP)  celle\n"
       << "du type Naturel (fait maison) pour des nombres de plus en plus grands." << endl;
  Naturel a,b,c,d;
  Integer x,y,z;
  Timer timer;

  cout << endl << "+++ Addition +++" << endl;
  for( unsigned long bits=1000; bits<=256000; bits*=2 )
    {
      //-----------------------------------------------------------------
      // Choix des nombres alatoirs et affichage de la taille
      //-----------------------------------------------------------------
      cout << "construction...\r" << flush;
      x= random(bits);
      y= random(bits);
      cout << "conversion...  \r" << flush;
      affecter(a,x);
      affecter(b,y);
      cout << setw(8) << mpz_sizeinbase( x.get_mpz_t(), 2 ) << " bits / " 
	   << setw(7) << a.size() << " dcimales";
      
      //-----------------------------------------------------------------
      // Chronomtrer l'addition GMP vs maison
      //-----------------------------------------------------------------
      cout << " : GMP " << flush;
      timer.start();
      z= x+y;
      timer.stop();
      cout << timer << ", maison " << flush;
      timer.start();
      c= a+b;
      timer.stop();
      cout << timer << flush;

#ifdef __verification__
      // Vrifier si les rsultats concident (juste pour tre sr)
      affecter(d,z);
      if ( c == d ) cout << ", ok"; else cout << ", ERREUR !";
#endif // __verification__
      cout << endl;
    };

  cout << endl << "*** Multiplication ***" << endl;
  double time_gmp= 0, time_karatsuba= 0, time_scolaire= 0;
  for ( unsigned long bits=1000; bits<=100000; bits+=1000 )
    {
      //-----------------------------------------------------------------
      // Choix des nombres alatoires et affichage de la taille
      //-----------------------------------------------------------------
      cout << "construction...\r" << flush;
      x= random(bits);
      y= random(bits);
      cout << "conversion...  \r" << flush;
      affecter(a,x);
      affecter(b,y);
      cout << setw(8) << mpz_sizeinbase( x.get_mpz_t(), 2 ) << " bits / " 
	   << setw(7) << a.size() << " dcimales";

      //-----------------------------------------------------------------
      // Chronomtrer la multiplication GMP
      //-----------------------------------------------------------------
      cout << " : GMP " << flush;
      timer.start();
      z= x*y;
      timer.stop();
      cout << timer;
      time_gmp= timer.difftime();
      if ( time_gmp > 10 ) { cout << endl; break; }
      
      //-----------------------------------------------------------------
      // Chronomtrer Karatsuba maison (si ce n'est pas trop lent)
      //-----------------------------------------------------------------
      if ( time_karatsuba > 10 ) { cout << endl; continue; }
      cout << ", Karatsuba maison " << flush;
      timer.start();
      karatsuba(a,b,c);
      timer.stop();
      cout << timer << flush;
      time_karatsuba= timer.difftime();

#ifdef __verification__
      // Vrifier si les rsultats concident (juste pour tre sr)
      affecter(d,z);
      if ( c == d ) cout << ", ok"; else cout << ", ERREUR !";
#endif // __verification__

      //-----------------------------------------------------------------
      // Chronomtrer la multiplication scolaire (si ce n'est pas trop lent)
      //-----------------------------------------------------------------
      if ( time_scolaire > 10 ) { cout << endl; continue; }
      cout << ", scolaire maison " << flush;
      timer.start();
      multiplication(a,b,c);
      timer.stop();
      cout << timer << flush;
      time_scolaire= timer.difftime();

#ifdef __verification__
      // Vrifier si les rsultats concident (juste pour tre sr)
      if ( c == d ) cout << ", ok"; else cout << ", ERREUR !";
#endif // __verification__
      cout << endl;
    };
}
