//=======================================================================
// timer.hh : horloge pour chronomtrer l'excution de logiciels en C++
// Copyright (C) 2005-2008 Michael Eisermann, all rights reserved
// Institut Fourier, Universit Joseph Fourier, Grenoble, France
//
// Description :
// La classe Timer permet de mesurer une dure de temps. L'horloge est 
// initialise par start() et arrte par stop(), puis l'affichage donne
// le temps coul en secondes et en ticks (= cycles processeur). 
//
// Remarque : 
// La prcision de l'horloge  clock()  est de 0.01s typiquement.
// Ainsi il est plus prcis et parfois plus informatif de compter 
// le nombre de cycles processeur.  Ainsi, pour les courtes dures
// de moins d'une seconde, on affiche aussi le nombre de ticks.
//
// Pour plus d'informations vous pouvez taper, en ligne de commande, 
//   $ cat /proc/cpuinfo | grep "cpu MHz" 
// Cette commande affiche la vitesse de votre ordinateur en MHz.
//=======================================================================

#ifndef _TIMER_HH_
#define _TIMER_HH_

#include <iostream>        // entre-sortie standard
#include <ctime>           // fonctions horloge du langage C
using namespace std;       // accs direct aux fonctions standards

class Timer
{
public:
  // Type pour compter les cycles processeur (ticks)
  typedef unsigned long long int Ticks;
  
private:
  // Dbut et fin de la dure mesure
  clock_t clock_start, clock_stop;
  Ticks ticks_start, ticks_stop;

  // Dterminer le temps coul (ticks = nombre de cycles processeur).
  // RDTSC est la commande assembleur pour `Read Time-Stamp Counter'.
  Ticks get_ticks()
    {
      unsigned long long int t;
      asm volatile ( "RDTSC" : "=A" (t) );
      return t;
    };
  
public:
  // Constructeur (initialiser l'horloge)
  Timer()
    : clock_start(clock()), 
      clock_stop(clock_start),
      ticks_start(get_ticks()), 
      ticks_stop(ticks_start) {};

  // Initialiser l'horloge
  void start()
    {
      clock_stop= clock_start= clock();
      ticks_stop= ticks_start= get_ticks();
    };

  // Arrter l'horloge
  void stop()
    {
      ticks_stop= get_ticks();
      clock_stop= clock();
    };

  // Dterminer le temps coul en secondes 
  double difftime() const
    {
      return double(clock_stop-clock_start) / CLOCKS_PER_SEC;
    };

  // Dterminer le temps coul en ticks
  Ticks ticks() const
    {
      return ticks_stop - ticks_start;
    };

  // Afficher le temps coul en secondes et ticks
  friend ostream& operator<< ( ostream& out, const Timer& timer )
    {
      // Formater le flot, tout en sauvegardant l'ancien format
      ios::fmtflags old_flags= out.flags();
      int old_precision= out.precision(2);
      out.setf( ios::fixed );
      
      // Afficher puis restituer l'ancien format du flot
      double secondes= timer.difftime(); 
      out << secondes << "s"; 
      if ( secondes < 1.0 ) 
	cout << " (" << timer.ticks() << " ticks)";
      out.precision( old_precision );
      out.flags( old_flags );
      return out;
    };
};

#endif /* _TIMER_HH_ */
