//=======================================================================
// temoins.cc : petite statistique des tmoins contre la primalit
// Copyright (C) 2004-2007 Michael Eisermann, all rights reserved
// Institut Fourier, Universit Joseph Fourier, Grenoble, France
//=======================================================================

#include <iostream>
#include <vector>
#include <string>
using namespace std;

typedef long long int Integer;
const int kCalculeMax= 1000000;
const size_t kAfficheMax= 20;

//-----------------------------------------------------------------------
// Calculer le pgcd par l'algorithme d'Euclide
// Entre : deux entier a et b
// Sortie : le pgcd de a et b
//-----------------------------------------------------------------------
Integer pgcd( Integer a, Integer b )
{
  while ( b != 0 )
    {
      Integer r= a%b;
      a= b; 
      b= r;
    };
  return abs(a);
}

//-----------------------------------------------------------------------
// Tester l'inversibilit de x modulo n
// Entre : un entier n et un entier x dans l'intervalle [1,n-1]
// Sortie : true si x est inversible modulo n, false sinon
//-----------------------------------------------------------------------
bool inversible( Integer n, Integer x )
{
  return ( pgcd(n,x) == 1 );
}

//-----------------------------------------------------------------------
// Calculer base^exp modulo mod par une puissance dichotomique
// Entre : un entier base, un entier exp >= 0, un entier mod != 0
// Sortie : un entier puiss congru  base^exp modulo mod
//-----------------------------------------------------------------------
Integer puissance( Integer base, Integer exp, const Integer& mod )
{
  Integer puiss(1);
  for( base%= mod ; exp > 0 ; puiss*= base, puiss%= mod, --exp )
    while( exp%2 == 0 ) { base*= base; base%= mod; exp/= 2; };
  return puiss;
}

//-----------------------------------------------------------------------
// Le test de Fermat modulo n
// Entre : un entier impair n et un entier x dans l'intervalle [1,n-1]
// Sortie : true si x^(n-1) = 1 modulo n, false sinon
//-----------------------------------------------------------------------
bool testFermat( Integer n, Integer x )
{
  return ( puissance( x, n-1, n ) == 1 );
}

//-----------------------------------------------------------------------
// Le test de racine modulo n
// Entre : un entier impair n et un entier x dans l'intervalle [1,n-1]
// Sortie : true si x^[(n-1)/2] = 1 modulo n, false sinon
//-----------------------------------------------------------------------
bool testRacine( Integer n, Integer x )
{
  Integer r= puissance( x, (n-1)/2, n );
  return ( r==1 || r==n-1 );
}

//-----------------------------------------------------------------------
// Le test de Miller-Rabin modulo n
// Entre : un entier impair n et un entier x dans l'intervalle [1,n-1]
// Sortie : true si n est pseudo-premier  base x, false sinon
//-----------------------------------------------------------------------
bool testMillerRabin( Integer n, Integer x )
{
  // prparer les paramtres et attraper des cas exceptionnels
  if ( n < 0 ) n= -n;
  if ( n <= 1 ) return false;
  if ( n <= 3 ) return true; 
  x%= n;
  if ( x<0 ) x+= n;

  // trouver la dcomposition  n-1 = q2^e
  int e= 0;
  Integer q= n-1;
  while ( q%2 == 0 ) { q/=2; ++e; }
  
  // effectuer le test de Miller-Rabin
  x= puissance( x, q, n );
  if ( x == 1 || x == n-1 ) return true;
  for ( --e ; e>0 ; --e ) 
    {
      x*= x;
      x%= n;
      if ( x == n-1 ) return true;
    };
  return false;
}

//-----------------------------------------------------------------------
// Afficher un vecteur (en bornant la longueur maximum)
//-----------------------------------------------------------------------
void affiche( const vector<Integer>& vec )
{
  size_t borne= ( vec.size() < kAfficheMax ? vec.size() : kAfficheMax );
  if ( borne == 0 ) return;
  cout << "( ";
  for ( size_t i=0; i<borne; ++i ) cout << vec[i] << " ";
  if ( kAfficheMax < vec.size() ) cout << "...";
  cout << ")" << endl;
}

//-----------------------------------------------------------------------
// Statistique des tmoins 
//-----------------------------------------------------------------------
void temoins( Integer n )
{
  vector<Integer> noninversibles, nonfermat, nonracine, nonmiller, passe;
  for( Integer a=1; a<n; ++a )
    {
      // tester d'abord si a est inversible modulo n
      if ( !inversible(n,a) )
	{
	  noninversibles.push_back(a);
	  continue;
	}

      // tester le petit thorme de Fermat
      if ( !testFermat(n,a) )
	{
	  nonfermat.push_back(a);
	  continue;
	}

      // tester la racine de petit Fermat
      if ( !testRacine(n,a) )
	{
	  nonracine.push_back(a);
	  continue;
	}

      // tester la racine de petit Fermat
      if ( !testMillerRabin(n,a) )
	{
	  nonmiller.push_back(a);
	  continue;
	}

      passe.push_back(a);
    };

  if ( noninversibles.size() == 0 )
    {
      cout << "Tous les lments non nuls sont inversibles." << endl
	   << "Ceci veut dire que " << n << " est premier." << endl;
      return;
    };
  
  cout << "Parmi les " << n-1 << " lments non nuls, " << noninversibles.size()
       << " ne sont pas inversibles." << endl;
  affiche( noninversibles );
  
  cout << "Parmi les lments restants, " << nonfermat.size() 
       << " ne satisfont pas au critre de Fermat." << endl;
  affiche( nonfermat );
  
  cout << "Parmi les lments restants, " << nonracine.size()
       << " ne satisfont pas au critre de racine." << endl;
  affiche( nonracine );
  
  cout << "Parmi les lments restants, " << nonmiller.size()
       << " ne satisfont pas au critre de Miller-Rabin." << endl;
  affiche( nonmiller );
  
  cout << "Au total, seulement " << passe.size() 
       << " lments passent tous les tests (soit moins de " 
       << (100*passe.size()+n-2)/(n-1) << "%)." << endl;
  affiche( passe );
}

//-----------------------------------------------------------------------
// La fonction main gre la communication avec l'utilisateur
//-----------------------------------------------------------------------
int main()
{
  cout << "\nBienvenue  la recherche des tmoins !" << endl;
  cout << "Pour un *petit* nombre naturel impair n donn, ce programme explicite les lments dans Z/n\n"
       << "qui sont non inversibles, ceux qui ne satisfont pas au test de Fermat x^(n-1) = 1,\n"
       << "et ceux qui ne satisfont pas au test x^[(n-1)/2] = 1. Dans un tel cas on dit que\n"
       << "x tmoigne que n est compos, o encore que x tmoigne contre la primalit de n.\n"
       << "Le test de Miller-Rabin en est une version optimise.\n";
  for(;;)
    {
      cout << "\nEntrez un petit nombre naturel impair ou 0 pour terminer : ";
      Integer n;
      cin >> n;
      if ( n == 0 ) break;
      if ( n%2 == 0 )
	{
	  cout << "Ce nombre est pair." << endl;
	  continue;
	};
      if ( n == 1 || n == -1 )
	{
	  cout << "Les lments 1 sont les units dans l'anneau Z." << endl;
	  continue;
	};
      if ( abs(n) > kCalculeMax ) 
	{
	  cout << "Ce nombre est trop grand." << endl;
	  continue;
	};
      if ( n < 0 )
	{
	  cout << "Ce nombre est ngatif; je prendrai la valeur absolue..." << endl;
	  n= -n;
	};
      temoins(n);
    };
  cout << "Au revoir et  bientt.\n" << endl;
}
