/*=======================================================================*\
 * fourier.cc : transformation de Fourier rapide (FFT)
 * Copyright (C) 2004 by Michael Eisermann, all rights reserved
 * Institut Fourier, Universit Joseph Fourier, Grenoble, France
 *-----------------------------------------------------------------------
 * Description :
 *-----------------------------------------------------------------------
 * Implmentation :
 *-----------------------------------------------------------------------
 * Compilation :
 *  compiler avec l'option -O3 pour optimiser un peu la rapidit. 
\*=======================================================================*/

#include <iostream>           // dclarer l'entre-sortie standard
#include <iomanip>
#include <vector>             // dfinir la classe gnrique vector
#include <cstdlib>            // pour dclarer random()
#include <ctime>              // dclarer clock() pour le chronomtrage
#include "pecoprem.hh"
using namespace std;          // accs direct aux fonctions standard

typedef int Indice;

const signed long long int p= 167772161;  // p = 5 2^25 + 1 est premier
const GF<p> primitive(3);                 // 3 est une racine primitive
const GF<p> omega(243);                   // 3^5 est d'ordre 2^25

/*-----------------------------------------------------------------------*\
 * Transformation de Fourier lente
 * Hypothse : soit K un corps et omega un lment d'ordre n dans K
 * Entre    : un vecteur a de longueur n, reprsentant le polynme A
 * Sortie    : le vecteur b = ( A(1), A(omega), ... , A(omega^n-1) )
\*-----------------------------------------------------------------------*/
template <typename K>
void transFourierLente( const vector<K>& a, vector<K>& b, const K& omega )
{
  const Indice n= a.size(); // dterminer la longueur n 
  b.clear();                // effacer l'ancien contenu du vecteur b 
  b.resize(n);              // rserver de la place dans le vecteur b
  Indice i,j;               // i parcourt dans a alors que j parcourt dans b
  K s, x= 1;                // s est une somme auxiliaire pour calculer A(x)
  for( Indice j=0; j<n; ++j )
    {
      // valuation de A(x) par la mthode de Horner
      for( s= a[n-1], i= n-2; i>=0; s*= x, s+= a[i], --i );
      b[j]= s;
      x*= omega;
    };
};
  
/*-----------------------------------------------------------------------*\
 * Transformation de Fourier rapide
\*-----------------------------------------------------------------------*/
template <typename K>
void transFourierRapide( const vector<K>& a, vector<K>& b, const K& omega )
{
  const Indice n= a.size();
  if ( n <= 1 )
    { 
      b=a; 
      return; 
    };

  if ( n%2 != 0 ) 
    { 
      transFourierLente(a,b,omega); 
      return; 
    };

  const Indice m= n/2;
  vector<K> c(m), d(m), e(m);
  for( Indice i=0; i<m; ++i ) 
    {
      c[i]= a[i]+a[i+m];
    };
  K omega2= omega * omega;
  transFourierRapide( c, d, omega2 );

  K x= 1;
  for( Indice i=0; i<m; ++i ) 
    {
      c[i]= ( a[i] - a[i+m] ) * x;
      x*= omega;
    };
  transFourierRapide( c, e, omega2 );
  
  b.clear();
  b.resize(n);
  for( Indice i=0, j=0; i<m; ++i, j+=2 )
    {
      b[j]= d[i];
      b[j+1]= e[i];
    };
}

template <typename K>
void multiplicationScolaire( const vector<K>& a, const vector<K>& b, vector<K>& c )
{
  while ( !a.empty() && zero(a.back()) ) a.popback();
  while ( !b.empty() && zero(b.back()) ) b.popback();
  c.clear();
  if ( a.empty() || b.empty() ) return;
  Indice m= a.size() + b.size() - 1;
  c.resize( m, K(0) );
  for( Indice i=0; i<a.size(); ++i )
    for( Indice j=0; j<b.size(); ++j )
      c[i+j]+= a[i]*b[j];
  while ( !c.empty() && zero(c.back()) ) c.popback();
}

template <typename K>
void multiplicationFFT( vector<K> a, vector<K> b, vector<K>& c, const K& omega, int exp )
{
  while ( !a.empty() && zero(a.back()) ) a.popback();
  while ( !b.empty() && zero(b.back()) ) b.popback();
  c.clear();
  if ( a.empty() || b.empty() ) return;
  if ( a.size() < 10 || b.size() < 10 )
    {
      multiplicationScolaire(a,b,c);
      return;
    };
  
  Indice m= a.size() + b.size() - 1;
  Indice n= 1, e= 0;
  while ( n<m ) { n*= 2, ++e; };
  if ( e>exp ) 
    {
      multiplicationScolaire(a,b,c);
      return;
    };

  a.resize( n, K(0) );
  b.resize( n, K(0) );
  K xi= omega;
  K dv= K(1)/K(n);
  for( Indice i= exp; exp>e; --i ) xi*= xi;
  
  vector<K> aa, bb, cc;
  transFourierRapide( a, aa, xi );
  transFourierRapide( b, bb, xi );
  for( Indice i= 0; i<n; ++i )
    cc[i]= (aa[i] * bb[i]) * dv;
  transFourierRapide( cc, c, 1/xi );
  while ( !c.empty() && zero(c.back()) ) c.popback();
}


void main()
{
  clock_t start, end;
  double time1= 0, time2= 0;
  cout.setf( ios::fixed );
  cout.precision(2);

  for( int n=1, e=0; e<=25; n*=2, ++e )
    {
      cout << "taille 2^" << setw(2) << e << " : " << flush;
      GF<p> r= omega;
      for( int i=25; i>e; --i ) r*= r;

      cout << "random... " << flush;
      vector< GF<p> > a(n), b, c;
      for( int i=0; i<n; ++i ) a[i]= random();
      
      cout << "Fourier rapide... " << flush;
      if ( time1 > 60 )
	{
	  cout << "abandon  " << flush;
	}
      else
	{
	  start= clock();
	  transFourierRapide(a,b,r);
	  end= clock();
	  time1= double( end-start ) / CLOCKS_PER_SEC;
	  cout << setw(6) << time1 << "s  " << flush;
	};

      cout << "Fourier lente... " << flush;
      if ( time2 > 60 )
	{
	  cout << "abandon " << flush;
	}
      else
	{
	  start= clock();
	  transFourierLente(a,c,r);
	  end= clock();
	  time2= double( end-start ) / CLOCKS_PER_SEC;
	  cout << setw(6) << time2 << "s  " << flush;
	  
	  cout << "vrification " << flush;
	  for( int i=0; i<n; ++i )
	    if ( b[i] != c[i] ) 
	      {
		cerr << endl << "Erreur ! rsultats incohrents." << endl;
		exit(0);
	      };  
	  cout << "ok";
	};
      
      cout << endl;
    }
}
