//=======================================================================
// integer.cc : arithmtique lmentaire des nombres entiers
// Copyright (C) 2007 by <votre nom>, all rights reserved
// 
// Description :
// Ce fichier implmente quelques fonctions de l'arithmtique 
// des nombres entiers : factorielle, coefficient binomial, puissance
// dichotomique (dans Z mais aussi modulo m), racine n-ime, etc ...
//
// Compilation :
//  compiler avec l'option -lgmpxx pour inclure la bibliothque GMP.
//
// Commentaire :
//  complter  fur et  mesure en rajoutant les solutions d'exercices.
//=======================================================================

#include <gmpxx.h>         // accder  la bibliothque GMP pour C++
typedef mpz_class Integer; // `Integer' semble plus parlant que `mpz_class'
using namespace std;       // accs direct aux fonctions standard

//=======================================================================
// Premire partie : dclarations / sommaire des fonctions arithmtiques
//=======================================================================

//-----------------------------------------------------------------------
// Fonctions de base pour un anneau effectif 
//-----------------------------------------------------------------------
inline bool    zero      ( const Integer& a );
inline bool    one       ( const Integer& a );
inline bool    inversible( const Integer& a );
inline bool    inversible( const Integer& a, Integer& b );
inline bool    divisible ( const Integer& a, const Integer& b );
inline bool    divisible ( const Integer& a, const Integer& b, Integer& q );
inline bool    associes  ( const Integer& a, const Integer& b );
inline bool    associes  ( const Integer& a, const Integer& b, Integer& u );
inline Integer assopref  ( const Integer& a );
inline Integer assopref  ( const Integer& a, Integer& u );

//-----------------------------------------------------------------------
// Les trois divisions euclidiennes les plus utiles
//-----------------------------------------------------------------------
void    tdiv( const Integer& a, const Integer& b, Integer& q, Integer& r );
Integer tdiv( const Integer& a, const Integer& b );
Integer tmod( const Integer& a, const Integer& b );

void    pdiv( const Integer& a, const Integer& b, Integer& q, Integer& r );
Integer pdiv( const Integer& a, const Integer& b );
Integer pmod( const Integer& a, const Integer& b );

void    sdiv( const Integer& a, const Integer& b, Integer& q, Integer& r );
Integer sdiv( const Integer& a, const Integer& b );
Integer smod( const Integer& a, const Integer& b );

Integer pgcd( Integer a, Integer b );
Integer pgcd( Integer a, Integer b, Integer& u );
Integer pgcd( Integer a, Integer b, Integer& u, Integer &v );

//-----------------------------------------------------------------------
// Fonctions diverses
//-----------------------------------------------------------------------
int     len( const Integer& x );
int     log2( const Integer& x );
Integer random( const Integer& min, const Integer& max );
Integer puissance( Integer base, Integer exp );
Integer puissance( Integer base, Integer exp, const Integer& mod );
Integer racine( Integer a, int n );
Integer factorielle( Integer n );
Integer binomial( Integer n, Integer k );
int     jacobi( Integer a, Integer b );

//=======================================================================
// Seconde partie : implmentations
//=======================================================================

//-----------------------------------------------------------------------
// Implmentation de quelques tests de base : reconnatre les lments
// 0 et 1, puis les lments inversibles, tester la divisibilit, puis
// regrouper les lments associs.
//-----------------------------------------------------------------------

// Reconnatre l'lment 0
inline bool zero( const Integer& a )
{ 
  return ( a == 0 ); 
}

// Reconnatre l'lment 1
inline bool one( const Integer& a )
{ 
  return ( a == 1 ); 
}

// Dterminer si a est inversible
inline bool inversible( const Integer& a )
{ 
  return ( a == 1 || a == -1 ); 
}

// Dterminer si a est inversible, si oui calculer son inverse
inline bool inversible( const Integer& a, Integer& b )
{ 
  if ( a == 1 || a == -1 ) { b=a; return true; } 
  else return false; 
}

// Dterminer si a est divisible par b
inline bool divisible( const Integer& a, const Integer& b )
{
  // Attraper le cas particulier: 0 divise 0 mais rien d'autre.
  if ( b == 0 ) return ( a == 0 );

  // Dterminer la divisibilit via la division euclidienne
  return ( a%b == 0 );
}

// Dterminer si a est divisible par b, si oui calculer q=a/b
inline bool divisible( const Integer& a, const Integer& b, Integer& q )
{
  // Attraper le cas particulier: 0 divise 0 mais rien d'autre.
  if ( b == 0 ) { if ( a == 0 ) { q=1; return true; } else return false; };

  // La formulation vidente calcule deux fois la mme division:
  // if ( a%b == 0 ) { q=a/b; return true; } else return false;
  // On peut l'optimiser pour n'effectuer qu'une seule division:
  Integer reste, quotient;
  tdiv( a, b, quotient, reste );
  if ( reste == 0 ) { q= quotient; return true; } else return false;
}

// Dterminer si a et b sont associs
inline bool associes( const Integer& a, const Integer& b )
{
  return ( a == b || a == -b );
}

// Dterminer si a et b sont associs, si oui calculer u inversible tel que ua=b
inline bool associes( const Integer& a, const Integer& b, Integer& u )
{
  if ( a == b )  { u=1;  return true; };
  if ( a == -b ) { u=-1; return true; };
  return false;
}

// Calculer l'associ prfr (ici : l'associ non-ngatif)
inline Integer assopref( const Integer& a )
{
  if ( a >= 0 ) return a; else return -a;
}

// Calculer l'associ prfr (ici : l'associ non-ngatif) et le facteur u
inline Integer assopref( const Integer& a, Integer& u )
{
  if ( a >= 0 ) { u=1; return a; } else { u=-1; return -a; };
}

//-----------------------------------------------------------------------
// Implmentation d'une division euclidienne de deux entiers a et b.
// On cherche donc q et r de sorte que a == q*b + r et abs(r) < abs(b).
// Le standard C++ prvoit que le quotient q soit obtenu par troncature,
// c'est--dire on choisit (q,r) de sorte que |q| soit minimal.
//-----------------------------------------------------------------------
void tdiv( const Integer& a, const Integer& b, Integer& q, Integer& r )
{
  mpz_tdiv_qr( q.get_mpz_t(), r.get_mpz_t(), a.get_mpz_t(), b.get_mpz_t() );
}
 
Integer tdiv( const Integer& a, const Integer& b )
{ 
  return a/b;
}

Integer tmod( const Integer& a, const Integer& b )
{
  if ( b == 0 ) return a;
  return a%b;
}

//-----------------------------------------------------------------------
// Implmentation d'une division euclidienne de deux entiers a et b.
// On choisit ici q et r de sorte que a == q*b + r et 0 <= r < abs(b),
// autrement dit, c'est la division euclidienne avec reste positif.
//-----------------------------------------------------------------------
void pdiv( const Integer& a, const Integer& b, Integer& q, Integer& r )
{
  tdiv(a,b,q,r);
  if ( r<0 ) { if ( b>0 ) { r+=b; --q; } else { r-=b; ++q; }; };
}

Integer pdiv( const Integer& a, const Integer& b )
{
  Integer q,r;
  pdiv(a,b,q,r);
  return q;
}

Integer pmod( const Integer& a, const Integer& b )
{
  if ( b == 0 ) return a;
  Integer r= a%b;
  if ( r<0 ) r+= abs(b);
  return r;
}

//-----------------------------------------------------------------------
// Implmentation d'une division euclidienne de deux entiers a et b.
// On choisit ici q et r de sorte que a == q*b + r et 2*abs(r) <= abs(b),
// c'est la division euclidienne avec reste symtrique (= reste minimal).
//-----------------------------------------------------------------------
void sdiv( const Integer& a, const Integer& b, Integer& q, Integer& r )
{
  tdiv(a,b,q,r);
  if ( b>0 )
    {
      if ( 2*r > b )  { r-=b; ++q; return; };
      if ( 2*r < -b ) { r+=b; --q; return; };
    }
  else
    {
      if ( 2*r > -b ) { r+=b; --q; return; };
      if ( 2*r < b )  { r-=b; ++q; return; };
    }
}

Integer sdiv( const Integer& a, const Integer& b )
{
  Integer q,r;
  sdiv(a,b,q,r);
  return q;
}

Integer smod( const Integer& a, const Integer& b )
{
  if ( b == 0 ) return a;
  Integer r= a%b;
  if ( 2*r > abs(b) ) r-= abs(b);
  else if ( -2*r > abs(b) ) r+= abs(b); 
  return r;
}

//-----------------------------------------------------------------------
// Choix d'une division euclidienne prfre
//-----------------------------------------------------------------------
inline void eudiv( const Integer& a, const Integer& b, Integer& q, Integer& r )
{ pdiv(a,b,q,r); }

inline Integer eumod( const Integer& a, const Integer& b )
{ return pmod(a,b); }

inline Integer eudiv( const Integer& a, const Integer& b )
{ return pdiv(a,b); }

//-----------------------------------------------------------------------
// Calculer le pgcd par l'algorithme d'Euclide
// Entre : deux entier a et b
// Sortie : le pgcd de a et b
//-----------------------------------------------------------------------
Integer pgcd( Integer a, Integer b );

//-----------------------------------------------------------------------
// Calculer le pgcd par l'algorithme d'Euclide
// Entre : deux entier a et b
// Sortie : le pgcd de a et b avec u vrifiant ua = pgcd(a,b) modulo b
//-----------------------------------------------------------------------
Integer pgcd( Integer a, Integer b, Integer& u );

//-----------------------------------------------------------------------
// Calculer le pgcd par l'algorithme d'Euclide
// Entre : deux entier a et b
// Sortie : le pgcd de a et b avec u,v vrifiant ua+bv = pgcd(a,b) 
//-----------------------------------------------------------------------
Integer pgcd( Integer a, Integer b, Integer& u, Integer &v );

//-----------------------------------------------------------------------
// Dterminer la longueur en base 2 (version optimise pour GMP)
// Entre : un entier x (positif, ngatif ou nul)
// Sortie : 0 pour x=0  et  1+[log2|x|] pour |x|>0
//-----------------------------------------------------------------------
int len( const Integer& x )
{ 
  return mpz_sizeinbase( x.get_mpz_t(), 2 ); 
}

//-----------------------------------------------------------------------
// Dterminer le logarithme en base 2 (version optimise pour GMP)
// Entre : un entier x (positif, ngatif ou nul)
// Sortie : 0 pour x=0  et  [log2|x|] pour |x|>0
//-----------------------------------------------------------------------
int log2( const Integer& x )
{ 
  return mpz_sizeinbase( x.get_mpz_t(), 2 ) - 1; 
}

//-----------------------------------------------------------------------
// Construire un nombre entier alatoire
// Entre : deux entiers min <= max
// Sortie : un entier dans l'intervalle [min,max]
//-----------------------------------------------------------------------
Integer random( const Integer& min, const Integer& max )
{
  static gmp_randclass generateur(gmp_randinit_default);
  return generateur.get_z_range(max-min+1) + min;
}

//-----------------------------------------------------------------------
// Calculer base^exp par une puissance dichotomique
// Entre : un entier base, un entier exp >= 0
// Sortie : l'entier base^exp
//-----------------------------------------------------------------------
Integer puissance( Integer base, Integer exp );

//-----------------------------------------------------------------------
// Calculer base^exp modulo mod par une puissance dichotomique
// Entre : un entier base, un entier exp >= 0, un entier mod != 0
// Sortie : un entier puiss quivalent  base^exp modulo mod
//-----------------------------------------------------------------------
Integer puissance( Integer base, Integer exp, const Integer& mod );

//-----------------------------------------------------------------------
// Calculer la partie entire de la racine n-ime
// Entre : deux entiers non-ngatifs a et n
// Sortie : l'unique entier non-ngatif r vrifiant  r^n <= a < (r+1)^n
//-----------------------------------------------------------------------
Integer racine( Integer a, int n );

//-----------------------------------------------------------------------
// Calculer la factorielle n! d'un nombre naturel n
// Entre : un entier n
// Sortie : la factorielle n! ou bien 0 si n<0
//-----------------------------------------------------------------------
Integer factorielle( Integer n );

//-----------------------------------------------------------------------
// Calculer le coefficient binomial (n,k)
// Entre : deux entiers n et k (typiquement avec 0 <= k <= n) 
// Sortie : le coefficient binomial (n,k), c'est--dire le nombre de
//          sous-ensembles de cardinal k d'un ensemble de cardinal n.
//-----------------------------------------------------------------------
Integer binomial( Integer n, Integer k );

//-----------------------------------------------------------------------
// Calculer le symbole de Jacobi
// Entre : deux entier a,b avec b impair
// Sortie : le symbole de Jacobi de (a,b)
//-----------------------------------------------------------------------
int jacobi( Integer a, Integer b );

//-----------------------------------------------------------------------
// Fin du fichier integer+.cc
//-----------------------------------------------------------------------

