//=======================================================================
// pecoprem.hh : classe pour modliser les petits corps premiers
// Copyright (C) 2004-2007 by Michael Eisermann, all rights reserved
// Institut Fourier, Universit Joseph Fourier, Grenoble, France
//-----------------------------------------------------------------------
// Description :
// La classe GF<p> (abrviation pour "galois field of order p") modlise 
// un corps premier, c'est--dire l'anneau quotient Z modulo p. Pour ceci
// on exige que p soit un petit nombre premier positif. La classe GF<2>
// est implmente sparment plus bas pour profiter des optimisations.
//-----------------------------------------------------------------------
// Restrictions :
// Pour stocker les valeurs on utilise le type SmInt ("small integer")
// qui doit tre un type primitif sign.  Pour la correction de GF<p> 
// on doit exiger que p soit un nombre premier et que p n'excde pas
// la capacit du type choisi. Cette restriction garantit que tous 
// les calculs, en particulier la multiplication, peuvent tre 
// effectus correctement avec le type SmInt. On a donc :
//      short int (16 bits) : 2 <= p <=        181 assure p < 2^15
// int = long int (32 bits) : 2 <= p <=      46337 assure p < 2^31
//  long long int (64 bits) : 2 <= p <= 3037000493 assure p < 2^63
// (La borne maximale est le plus grand premier tel que p<2^bits.)
//-----------------------------------------------------------------------
// Remarque :
// Si p n'est pas premier, alors toutes les oprations arithmtiques
// fonctionnent correctement sauf bien sr la division par un lment 
// non inversible. Dans ce cas 1/a renvoie 0 pour signaler l'erreur.
// Les fonctions inversible(a), divisible(a,b) et associes(a,b) supposent
// toutes que p est premier mais elles ne le testent videmment pas.
// Elles ne fonctionnent donc plus correctement dans le cas d'un 
// paramtre p qui soit compos.
//=======================================================================

#ifndef __pecoprem_hh__
#define __pecoprem_hh__

#include <iostream>
using namespace std;

// typedef signed short SmInt; 
typedef signed int   SmInt; 
// typedef signed long long int SmInt; 

template <SmInt p>
class GF
{
private:
  SmInt valeur;
  
  // Calculer le reprsentant canonique de x modulo p
  SmInt reste( SmInt x ) const
    { x%= p; if ( x<0 ) x+= p; return x; };

  // Calculer l'inverse de a sous condition que pgcd(a,p)==1
  SmInt inverse( SmInt x ) const
    { 
      SmInt a(reste(x)), b(p), u(1), v(0), q, r;
      // La boucle suivante calcule pgcd(a,b) == pgcd(x,p)
      // en prservant la proprit que p divise a-u*x et b-p*v
      while ( b!=0 ) 
	{ 
	  q= a/b; r= a%b;  a= b; b= r;
	  r= reste(u-q*v); u= v; v= r;
	}
      // Finalement a == pgcd(a,b) == pgcd(x,p) == 1,
      // donc u*x == 1 modulo p, comme souhait.
      if ( a==1 ) return u; else return 0;
    };
  
public:
  // Constructeurs
  GF( SmInt v=SmInt(0) ) 
    : valeur( reste(v) ) {};
  GF( const GF& a )    
    : valeur( a.valeur ) {};
  
  // Accder  la valeur (conversion de type)
  SmInt val() const
    { return valeur; };
  
  // Affectation d'une valeur
  GF& operator= ( SmInt v )
    { valeur= reste(v); return *this; };
  GF& operator= ( GF a )
    { valeur= a.valeur; return *this; };
  
  // Comparaison
  bool operator== ( GF a ) const
    { return ( valeur == a.valeur ); }; 
  bool operator!= ( GF a ) const
    { return ( valeur != a.valeur ); };
  
  // Addition
  GF operator+ ( GF a ) const
    { return GF( valeur + a.valeur ); };
  GF& operator+= ( GF a )
    { valeur= reste( valeur + a.valeur ); return *this; };
  
  // Ngation et soustraction
  GF operator- () const
    { return GF( -valeur ); };
  GF operator- ( GF a ) const
    { return GF( valeur - a.valeur ); };
  GF& operator-= ( GF a )
    { valeur= reste( valeur - a.valeur ); return *this; };
  
  // Incrmentation et dcrmentation
  GF& operator++ ()
    { ++valeur; if ( valeur>=p ) valeur=0; return *this; };
  GF& operator-- ()
    { --valeur; if ( valeur<0 ) valeur=p-1; return *this; };
  
  // Multiplication
  GF operator* ( GF a ) const
    { return GF( valeur * a.valeur ); };
  GF& operator*= ( GF a )
    { valeur= reste( valeur * a.valeur ); return *this; };

  // Division
  GF operator/ ( GF a ) const
    { return GF( valeur * inverse(a.valeur) ); };
  GF& operator/= ( GF a )
    { valeur= reste( valeur * inverse(a.valeur) ); return *this; };
};

//-----------------------------------------------------------------------
// Implmentation des oprateurs d'entre-sortie
//-----------------------------------------------------------------------
template <SmInt p>
ostream& operator<< ( ostream& out, const GF<p>& a )
{
  return out << a.val();
}

template <SmInt p>
istream& operator>> ( istream& in, GF<p>& a )
{
  SmInt v;
  in >> v;
  a= v;
  return in;
}

//-----------------------------------------------------------------------
// Implmentation de quelques oprateurs commodes (conversion de type)
//-----------------------------------------------------------------------
template <SmInt p> inline
bool operator== ( GF<p> a, SmInt b )
{ return a == GF<p>(b); }

template <SmInt p> inline
bool operator== ( SmInt a, GF<p> b )
{ return GF<p>(a) == b; }

template <SmInt p> inline
bool operator!= ( GF<p> a, SmInt b )
{ return a != GF<p>(b); }

template <SmInt p> inline
bool operator!= ( SmInt a, GF<p> b )
{ return GF<p>(a) != b; }

template <SmInt p> inline
GF<p> operator+ ( GF<p> a, SmInt b )
{ return a + GF<p>(b); }

template <SmInt p> inline
GF<p> operator+ ( SmInt a, GF<p> b )
{ return GF<p>(a) + b; }

template <SmInt p> inline
GF<p> operator- ( GF<p> a, SmInt b )
{ return a - GF<p>(b); }

template <SmInt p> inline
GF<p> operator- ( SmInt a, GF<p> b )
{ return GF<p>(a) - b; }

template <SmInt p> inline
GF<p> operator* ( GF<p> a, SmInt b )
{ return a * GF<p>(b); }

template <SmInt p> inline
GF<p> operator* ( SmInt a, GF<p> b )
{ return GF<p>(a) * b; }

template <SmInt p> inline
GF<p> operator/ ( GF<p> a, SmInt b )
{ return a / GF<p>(b); }

template <SmInt p> inline
GF<p> operator/ ( SmInt a, GF<p> b )
{ return GF<p>(a) / b; }

//-----------------------------------------------------------------------
// Fonctions requises pour travailler avec l'anneau GF<p>
//-----------------------------------------------------------------------
// Reconnatre l'lment 0
template <SmInt p> inline
bool zero( GF<p> a )
{ 
  return a.val() == 0; 
}

// Reconnatre l'lment 1
template <SmInt p> inline
bool one( GF<p> a )
{ 
  return a.val() == 1; 
}

// Dterminer si a est inversible
template <SmInt p> inline
bool inversible( GF<p> a )
{ 
  return !zero(a); 
}

// Dterminer si a est inversible, si oui calculer son inverse
template <SmInt p> inline
bool inversible( GF<p> a, GF<p>& inverse )
{ 
  if ( zero(a) ) return false; 
  inverse= 1/a; 
  return true; 
}

// Dterminer si a est divisible par b
template <SmInt p> inline
bool divisible( GF<p> a, GF<p> b )
{ 
  return zero(a) || !zero(b); 
}

// Dterminer si a est divisible par b, si oui calculer le quotient q
template <SmInt p> inline
bool divisible( GF<p> a, GF<p> b, GF<p>& q )
{ 
  if ( zero(a) ) { q= 0; return true; };
  if ( !zero(b) ) { q= a/b; return true; };
  return false;
}

// Dterminer si a et b sont associs
template <SmInt p> inline
bool associes( GF<p> a, GF<p> b )
{ 
  return ( zero(a) && zero(b) ) || ( !zero(a) && !zero(b) ); 
}

// Dterminer si a et b sont associs, si oui calculer u tel que ua=b
template <SmInt p> inline
bool associes( GF<p> a, GF<p> b, GF<p>& u )
{ 
  if ( zero(a) && zero(b) ) { u=1; return true; };
  if ( !zero(a) && !zero(b) ) { u= b/a; return true; };
  return false;
}

// Choisir un reprsentant prfr parmi les lments associs  a
template <SmInt p> inline
GF<p> assopref ( GF<p> a )
{
  if ( zero(a) ) return 0; else return 1;
}

// Choisir un reprsentant prfr r=ua parmi les lments associs  a
template <SmInt p> inline
GF<p> assopref ( GF<p> a, GF<p>& u )
{
  if ( zero(a) ) { u= 1; return 0; }
  else { u= 1/a; return 1; };
}

// Dterminer le pgcd de a et b (trivial dans un corps)
template <SmInt p> inline
GF<p> pgcd ( GF<p> a, GF<p> b )
{
  if ( zero(a) && zero(b) ) return 0;
  else return 1;
}

// Dterminer le pgcd et des coefficients u,v tels que  pgcd(a,b) = ua+vb.
template <SmInt p> inline
GF<p> pgcd ( GF<p> a, GF<p> b, GF<p>& u, GF<p>& v )
{
  if ( !zero(a) ) { u= 1/a; v= 0; return 1; };
  if ( !zero(b) ) { u= 0; v= 1/b; return 1; };
  u= 1; v= 1; return 0;
}

// Division euclidienne (tout corps est aussi un anneau euclidien)
template <SmInt p> inline
void eudiv ( GF<p> a, GF<p> b, GF<p>& q, GF<p>& r )
{
  q= a/b; r=0; return;
}

// Dterminer le quotient de la division euclidienne
template <SmInt p> inline
GF<p> eudiv ( GF<p> a, GF<p> b )
{
  return a/b;
}

// Dterminer le reste de la division euclidienne
template <SmInt p> inline
GF<p> eumod ( GF<p> a, GF<p> b )
{
  if ( zero(b) ) return a;
  else return GF<p>(0);
}

//=======================================================================
// Implmentation de la classe spcialise GF<2> modlisant le corps Z/2.
//-----------------------------------------------------------------------
// Une spcialisation de la classe gnrique GF<p> pour p==2 permet 
// d'optimiser l'usage de la mmoire et les oprations  effectuer :
// Les valeurs sont ainsi stockes par une variable de type bool 
// avec pour convention que false==0 et true==1, comme d'habitude.
// Une telle variable n'occupe qu'un seul bit, de plus les oprations 
// arithmtiques deviennent particulirement simples et rapides.
//=======================================================================

template<>
class GF<2>
{
public:
  bool valeur;
  
private:
  // Calculer le reprsentant canonique de x modulo 2
  bool reste( SmInt x ) const
    { return ( x%2 != 0 ); };
  
public:
  // Constructeurs
  GF( bool b=false ) 
    : valeur( b ) {};
  GF( const GF& a )    
    : valeur( a.valeur ) {};
  GF( int v ) 
    : valeur( reste(v) ) {};
  
  // Accder  la valeur par une conversion de type
  int val() const
    { if (valeur) return 1; else return 0; };
  
  // Affectation
  GF& operator= ( bool b )
    { valeur= b; return *this; };
  GF& operator= ( GF a )
    { valeur= a.valeur; return *this; };
  GF& operator= ( int v )
    { valeur= reste(v); return *this; };
  
  // Comparaison
  bool operator== ( GF a ) const
    { return ( valeur == a.valeur ); }; 
  bool operator!= ( GF a ) const
    { return ( valeur != a.valeur ); };
  
  // Addition
  GF operator+ ( GF a ) const
    { return GF( valeur ^ a.valeur ); };
  GF& operator+= ( GF a )
    { valeur^= a.valeur; return *this; };
  
  // Ngation et soustraction
  GF operator- () const
    { return *this; };
  GF operator- ( GF a ) const
    { return GF( valeur ^ a.valeur ); };
  GF& operator-= ( GF a )
    { valeur^= a.valeur; return *this; };
  
  // Incrmentation et dcrmentation
  GF& operator++ ()
    { valeur= !valeur; return *this; };
  GF& operator-- ()
    { valeur= !valeur; return *this; };
  
  // Multiplication
  GF operator* ( GF a ) const
    { return GF( valeur & a.valeur ); };
  GF& operator*= ( GF a )
    { valeur&= a.valeur; return *this; };

  // Division
  GF operator/ ( GF a ) const
    { return GF( valeur& a.valeur ); };
  GF& operator/= ( GF a )
    { valeur&= a.valeur; return *this; };
};

//-----------------------------------------------------------------------
// Implmentation de quelques oprateurs commodes (conversion de type)
//-----------------------------------------------------------------------

bool operator== ( GF<2> a, bool b )
{ return a.valeur == b; }

bool operator== ( bool a, GF<2> b )
{ return a == b.valeur; }

bool operator!= ( GF<2> a, bool b )
{ return a.valeur != b; }

bool operator!= ( bool a, GF<2> b )
{ return a != b.valeur; }

//-----------------------------------------------------------------------
// Fonctions requises pour travailler avec l'anneau GF<2>
//-----------------------------------------------------------------------
// Reconnatre l'lment 0
inline bool zero( GF<2> a ) 
{ 
  return !a.valeur; 
}

// Reconnatre l'lment 1
inline bool one( GF<2> a )
{ 
  return a.valeur; 
}

// Dterminer si a est inversible
inline bool inversible( GF<2> a )
{ 
  return a.valeur; 
}

// Dterminer si a est inversible, si oui calculer son inverse
inline bool inversible( GF<2> a, GF<2>& inverse )
{ 
  if ( !a.valeur ) return false; 
  inverse.valeur= true; 
  return true; 
}

// Dterminer si a est divisible par b
inline bool divisible( GF<2> a, GF<2> b )
{ 
  return !a.valeur || b.valeur; 
}

// Dterminer si a est divisible par b, si oui calculer le quotient q
inline bool divisible( GF<2> a, GF<2> b, GF<2>& q )
{ 
  if ( !a.valeur ) { q.valeur= 0; return true; };
  if ( b.valeur ) { q.valeur= a.valeur; return true; };
  return false;
}

// Dterminer si a et b sont associs
inline bool associes( GF<2> a, GF<2> b )
{ 
  return ( a.valeur == b.valeur );
}

// Dterminer si a et b sont associs, si oui calculer u tel que ua=b
inline bool associes( GF<2> a, GF<2> b, GF<2>& u )
{ 
  if ( a.valeur == b.valeur ) { u.valeur= 1; return true; };
  return false;
}

// Choisir un reprsentant prfr parmi les lments associs  a
inline GF<2> assopref( GF<2> a )
{
  return a;
}

// Choisir un reprsentant prfr r=ua parmi les lments associs  a
inline GF<2> assopref( GF<2> a, GF<2>& u )
{
  u.valeur= 1;
  return a;
}

// Dterminer le pgcd de a et b (trivial dans un corps)
inline GF<2> pgcd( GF<2> a, GF<2> b )
{
  if ( zero(a) && zero(b) ) return 0;
  else return 1;
}

// Dterminer le pgcd et des coefficients u,v tels que  pgcd(a,b) = ua+vb.
inline GF<2> pgcd( GF<2> a, GF<2> b, GF<2>& u, GF<2>& v )
{
  if ( !zero(a) ) { u= a; v= 0; return 1; };
  if ( !zero(b) ) { u= 0; v= b; return 1; };
  u= 1; v= 1; return 0;
}

// Division euclidienne (tout corps est aussi un anneau euclidien)
inline void eudiv( GF<2> a, GF<2> b, GF<2>& q, GF<2>& r )
{
  q= a/b; r=0; return;
}

// Dterminer le quotient de la division euclidienne
inline GF<2> eudiv( GF<2> a, GF<2> b )
{
  return a/b;
}

// Dterminer le reste de la division euclidienne
inline GF<2> eumod( GF<2> a, GF<2> b )
{
  if ( zero(b) ) return a;
  else return 0;
}

#endif // __pecoprem_hh__
