//=======================================================================
// polynome.hh : classe pour modliser les polynmes sur un anneau
// Copyright (C) 2004-2006 Michael Eisermann, all rights reserved
// Institut Fourier, Universit Joseph Fourier, Grenoble, France
//
// Description :
// tant donn un anneau A, la classe Polynome<A> modlise l'anneau 
// A[X] des polynmes  p = p[0] + p[1]X + ... + p[n]X^n  sur A.
//
// lments requis : 
// On exige que A implmente un anneau commutatif unitaire suivant 
// les spcifications donnes dans le fichier "anneau.cc".  Pour des
// fonctions plus sophistiques on exige de plus que A soit intgre,
// voire factoriel avec une fonction pgcd effective.
//
// lments fournis :
// La classe Polynome<A> implmente un anneau commutatif unitaire suivant 
// les mmes spcifications, modlisant l'anneau des polynmes sur A.
// Si A est intgre, alors A[X] est intgre.  Si A est factoriel, alors
// A[X] est factoriel, et la fonction pgcd sur A s'tend sur A[X].
//
// Implmentation : 
// Un polynme p[0] + p[1]X + ... + p[n]X^n  coefficients dans 
// un anneau A est stock sous forme d'un vecteur p index de 0  n. 
// On supprime systmatiquement tous les coefficients nuls terminaux.
// En particulier le polynme nul est reprsent par le vecteur vide, 
// et un polynme de degr n par un vecteur de taille n+1 exactement.
//
// Remarques :
// Les oprations usuelles comme affectation (=), comparaison (== et !=)
// et l'arithmtique (+,+= et -,-= et *,*=) ont la smantique usuelle. 
// Voici quelques points plus subtils ou moins standard :
//
// Coefficient dominant et degr d'un polynme
// - Tout polynme s'crit comme p[0] + ... + p[n]X^n avec p[n] non nul.
//   Dans ce cas on appelle p[n] le coefficient dominant et n le degr.
//   Pour le polynme nul il faut adopter une convention particulire.
//   Ici on convient que le degr du polynme nul vaut -1, et son
//   coefficient dominant vaut 0.
//
// Division euclidienne
// - Si a et b sont deux polynmes, et le coefficient dominant de b
//   est inversible, alors on peut effectuer la division euclidienne
//   de a par b comme d'habitude:  le rsultat est l'unique paire
//   q,r telle que a = qb + r avec deg(r) < deg(a). 
// - Les oprateurs / et % supposent que la division euclidienne soit
//   dfinie, et mettent un message d'erreur dans le cas contraire.
//
// Pseudo-division euclidienne
// - tant donn deux polynmes a et b on note c le coefficient dominant 
//   de b.  Si c vaut 1, ou si c est au moins inversible, alors on peut
//   effectuer la division euclidienne comme d'habitude.  Si c est non nul
//   seulement, il existe deux polynmes q et r et un entier k>=0 tels que 
//   c^ka = bq+r  avec  deg(r) < deg(b).  On choisira k minimal, bien sr,
//   et pour k=0 on retombe sur le cas de la vraie division euclidienne.
//   On implmente ici la version gnrale parce qu'elle s'applique  tout 
//   polynme non nul et permet une utilisation plus souple dans la suite.
//=======================================================================

#ifndef __polynome_hh__
#define __polynome_hh__

#include <iostream>
#include <vector>
#include <string>
#include <sstream>
using namespace std;

// Type global pour stocker le degr d'un polynme ou un indice
typedef int Degre;

// La classe Polynome<A> implmente les polynmes et leurs oprations.
template <typename A> class Polynome;
template <typename A> Degre deg( const Polynome<A>& p );
template <typename A> A dom( const Polynome<A>& p );
template <typename A> Polynome<A> operator+ ( Polynome<A> p, const Polynome<A>& q );
template <typename A> Polynome<A> operator- ( Polynome<A> p, const Polynome<A>& q );
template <typename A> Polynome<A> operator* ( const Polynome<A>& p, const Polynome<A>& q );
template <typename A> Polynome<A> operator^ ( Polynome<A> p, int n );

// Division euclidienne (avec dom(b) inversible)
template <typename A> void eudiv( const Polynome<A>& a, const Polynome<A>& b, Polynome<A>& q, Polynome<A>& r );
template <typename A> Polynome<A> eudiv( const Polynome<A>& a, const Polynome<A>& b );
template <typename A> Polynome<A> eumod( const Polynome<A>& a, const Polynome<A>& b );

// Pseudo-division euclidienne (avec dom(b) non nul)
template <typename A> int pseudo_division( const Polynome<A>& a, const Polynome<A>& b, Polynome<A>& q, Polynome<A>& r );
template <typename A> int pseudo_quotient( const Polynome<A>& a, const Polynome<A>& b, Polynome<A>& q );
template <typename A> int pseudo_reste( const Polynome<A>& a, const Polynome<A>& b, Polynome<A>& r );

// Fonctions de base requises pour travailler avec l'anneau Polynome<A>
template <typename A> bool zero( const Polynome<A>& p );
template <typename A> bool one( const Polynome<A>& p );
template <typename A> bool inversible( const Polynome<A>& p );
template <typename A> bool inversible( const Polynome<A>& p, Polynome<A>& q );
template <typename A> bool divisible( const Polynome<A>& p, const Polynome<A>& q );
template <typename A> bool divisible( const Polynome<A>& a, const Polynome<A>& b, Polynome<A>& c );
template <typename A> bool associes( const Polynome<A>& p, const Polynome<A>& q );
template <typename A> bool associes( const Polynome<A>& p, const Polynome<A>& q, A& u );
template <typename A> bool associes( const Polynome<A>& p, const Polynome<A>& q, Polynome<A>& u );
template <typename A> Polynome<A> assopref ( const Polynome<A>& p );
template <typename A> Polynome<A> assopref ( const Polynome<A>& p, A& u );
template <typename A> Polynome<A> assopref ( const Polynome<A>& p, Polynome<A>& u );

// Calculer le pgcd (sur un anneau factoriel A avec pgcd effectif)
template <typename A> A contenu( const Polynome<A>& a );
template <typename A> Polynome<A> pseudo_pgcd( Polynome<A> a, Polynome<A> b );

// Entre-sortie de polynmes
template <typename A> ostream& operator<< ( ostream& out, const Polynome<A>& p );
template <typename A> istream& operator>> ( istream& in, Polynome<A>& p );

//=======================================================================
// La classe gnrique Polynme<A> qui modlise les polynmes sur A
// Partie 1 : la dclaration de la classe
//=======================================================================
template <typename A>
class Polynome
{
public: 
  // Un synonyme public pour accder au type des coefficients
  typedef A Coeff;
  
private:
  // La liste de coefficients (vide par dfaut)
  vector<A> coeffs;
  
  // Supprimer d'ventuels zros superflus
  void normaliser()
    { while( !coeffs.empty() && zero(coeffs.back()) ) coeffs.pop_back(); };
  
public:
  //---------------------------------------------------------------------
  // Constructeurs divers
  //---------------------------------------------------------------------
  // Constructeur par copie d'un polynme p donn
  Polynome( const Polynome& p )
    : coeffs( p.coeffs ) {};
  
  // Crer un polynme  partir d'une liste de coefficients
  Polynome( const vector<A>& c )
    : coeffs( c ) { normaliser(); };
  
  // Constructeur par conversion d'un monme aX^n
  Polynome( int a=0, Degre n=0 )
    { if ( a!=0 && n>=0 ) 
      { coeffs.resize( n, A(0) ); coeffs.push_back( A(a) ); }; };
  
  // Constructeur par conversion d'un monme aX^n
  Polynome( const A& a, Degre n=0 )
    { if ( !zero(a) && n>=0 ) 
      { coeffs.resize( n, A(0) ); coeffs.push_back( a ); }; };
  
  // Crer un poylnme  partir d'une chane de caractres
  Polynome( const char* s )
    { istringstream in(s); in >> *this; };
  
  // Crer un polynme  partir d'un string
  Polynome( const string& s )
    { istringstream in(s.c_str()); in >> *this; };

  //---------------------------------------------------------------------
  // Affectations diverses
  //---------------------------------------------------------------------
  // Effacer (remettre  zro)
  void clear()
    { coeffs.clear(); };

  // change de deux polynmes
  void swap( Polynome& p )
    { using std::swap; swap( coeffs, p.coeffs ); }

  // Affectation d'un polynme
  Polynome& operator= ( const Polynome& p )
    { Polynome copie(p); swap(copie); return *this; };

  // Affectation d'un string
  Polynome& operator= ( const char* s )
    { Polynome copie(s); swap(copie); return *this; };

  // Affectation d'un string
  Polynome& operator= ( const string& s )
    { Polynome copie(s); swap(copie); return *this; };
  
  // Affectation d'un polynme
  Polynome& set( const Polynome& p )
    { Polynome copie(p); swap(copie); return *this; };
  
  // Affectation d'une liste de polynme
  Polynome& set( const vector<A>& c )
    { Polynome copie(c); swap(copie); return *this; };
  
  // Affectation d'un monme aX^n
  Polynome& set( int a, Degre n=0 )
    { Polynome copie(a,n); swap(copie); return *this; };
  
  // Affectation d'un monme aX^n
  Polynome& set( const A& a, Degre n=0 )
    { Polynome copie(a,n); swap(copie); return *this; };
  
  // Affecter un coefficient particulier par indice
  void set_coeff( Degre n, const A& a );
  void set_coeff( Degre n, int a ) { set_coeff( n, A(a) ); };
  
  //---------------------------------------------------------------------
  // Dterminer le degr et les coefficients
  //---------------------------------------------------------------------
  // Dterminer si le polynme est nul (= vecteur des coefficients vide)
  bool empty() const
    { return coeffs.empty(); }
  
  // Dterminer le degr (NB: degr n  <=>  vecteur de longueur n+1 !)
  Degre deg() const 
    { return coeffs.size() - 1; };

  // Dterminer le coefficient dominant
  const A& dom() const
    { if ( coeffs.empty() ) return nul; else return coeffs.back(); };
  
  // Lire les coefficients par indice
  const A& operator[] ( Degre n ) const
    { if ( n<0 || n>deg() ) return nul; else return coeffs[n]; };
  
  //---------------------------------------------------------------------
  // Comparaison
  //---------------------------------------------------------------------
  bool operator== ( const Polynome& p ) const
    { return ( coeffs == p.coeffs ); }; 

  bool operator== ( const A& a ) const
    { if ( coeffs.empty() ) return zero(a);
      return ( coeffs.size() == 1 && coeffs[0] == a ); };

  bool operator!= ( const Polynome& p ) const
    { return ( coeffs != p.coeffs ); }; 

  bool operator!= ( const A& a ) const
    { if ( coeffs.empty() ) return !zero(a);
      return ( coeffs.size() != 1 || coeffs[0] != a ); };
  
  //---------------------------------------------------------------------
  // Oprations arithmtiques
  //---------------------------------------------------------------------
  // Addition
  Polynome  operator+ () const;
  Polynome& operator+= ( const Polynome& p );
  Polynome& operator+= ( const A& a );
  
  // Oppos et soustraction
  Polynome  operator- () const;
  Polynome& operator-= ( const Polynome& p );
  Polynome& operator-= ( const A& a );
  
  // Multiplication
  Polynome& operator*= ( const A& a );
  Polynome& operator*= ( const Polynome& p )
    { coeffs= ( *this * p ).coeffs; return *this; };
  
  // Division euclidienne (le coefficient dominant de b doit tre inversible)
  friend void eudiv <>
  ( const Polynome<A>& a, const Polynome<A>& b, Polynome<A>& q, Polynome<A>& r );

  // Pseudo-division euclidienne (le coefficient dominant de b doit tre non nul)
  friend int pseudo_division <> 
  ( const Polynome<A>& a, const Polynome<A>& b, Polynome<A>& q, Polynome<A>& r );
  
  // Quotient de la division euclidienne (quand celle-ci est dfinie)
  Polynome operator/ ( const Polynome& b )
    { 
      Polynome q,r;
      if ( pseudo_division(*this,b,q,r) > 0 )
	cerr << "Avertissement : pseudo-division euclidienne !" << endl;
      return q;
    }

  // Quotient de la division euclidienne (oprateur en place)
  Polynome& operator/= ( const Polynome& p )
    { set( *this / p ); return *this; }
  
  // Reste de la division euclidienne (quand celle-ci est dfinie)
  Polynome operator% ( const Polynome& b )
    {
      Polynome q,r; 
      if ( pseudo_division(*this,b,q,r) > 0 ) 
	cerr << "Avertissement : pseudo-division euclidienne !" << endl;
      return r;
    }      

  // Reste de la division euclidienne (oprateur en place)
  Polynome& operator%= ( const Polynome& p )
    { set( *this % p ); return *this; }

  // Dcalage (multiplication par X^n, division euclidienne si n<0)
  void decaler( Degre n )
    {
      if ( n>=0 ) { coeffs.insert( coeffs.begin(), n, A(0) ); return; };
      if ( coeffs.size() + n <= 0 ) { coeffs.clear(); return; };
      coeffs.erase( coeffs.begin(), coeffs.begin() - n );
    }
  
  // Division d'un polynme par une constante
  Polynome& operator/= ( const A& a );

  // valuation du polynme (substituer X par une constante a)
  A operator() ( const A& a ) const;

  // valuation du polynme (substituer X par un polynme p)
  Polynome operator() ( const Polynome& p ) const;

  //---------------------------------------------------------------------
  // Oprateurs de l'entre-sortie 
  //---------------------------------------------------------------------
  // Entre-sortie
  friend ostream& operator<< <>( ostream& out, const Polynome<A>& p );
  friend istream& operator>> <>( istream& in, Polynome<A>& p );

private:
  // Dfinition du symbole qui reprsentera la variable lors de l'affichage
  static char leSymbole;
  
public:
  static char symbole()
    { return leSymbole; };
  static char symbole( char s )
    { char ancien= leSymbole; leSymbole= s; return ancien; };

private:
  // Dfinition d'un coefficient nul comme constante statique.
  // Cette astuce est ncessaire pour une simple raison technique : 
  // pour pouvoir renvoyer la valeur nulle sous forme d'une rfrence.
  static const A nul;
};

//=======================================================================
// Partie 2 : l'implmentation des fonctions
//=======================================================================

//-----------------------------------------------------------------------
// Variables statiques : ces variables sont partages par tous les
// objets de la classe, en l'occurrence les polynmes sur l'anneau A.
//-----------------------------------------------------------------------
template <typename A>
char Polynome<A>::leSymbole('X');

template <typename A>
const A Polynome<A>::nul(0);

//-----------------------------------------------------------------------
// Quelques reformulations (redondantes mais commodes)
//-----------------------------------------------------------------------
template<typename A>
void swap( Polynome<A>& p, Polynome<A>& q ) 
{ p.swap(q); }

template <typename A>
Degre deg( const Polynome<A>& p )
{ return p.deg(); } 

template <typename A>
const A& dom( const Polynome<A>& p )
{ return p.dom(); } 

//-----------------------------------------------------------------------
// Comparaison : deux oprateurs complmentaires 
// pour comparer un polynme et un scalaire
//-----------------------------------------------------------------------
template <typename A>
bool operator== ( const A& a, const Polynome<A>& p )
{ return p == a; }

template <typename A>
bool operator!= ( const A& a, const Polynome<A>& p )
{ return p != a; }

//-----------------------------------------------------------------------
// Affecter un coefficient
//-----------------------------------------------------------------------
// Cette tche a l'aire triviale, mais ce n'est pas tout  fait vrai.
// On doit prendre soin ici de ne pas stocker un coefficient dans une
// case qui n'existe pas encore dans le vecteur coeffs.  De plus il faut
// veiller  ce que le degr peut changer ; on doit ainsi assurer que le
// vecteur modifi sera  nouveau normalis.
//-----------------------------------------------------------------------
template <typename A>
void Polynome<A>::set_coeff( Degre n, const A& a )
{ 
  // Premier cas : l'indice est hors domaine, il n'y a rien  faire.
  if ( n<0 ) return; 

  // Deuxime cas : l'indice est plus petit que le degr, c'est facile.
  // Il suffit ici de stocker la valeur dans la case dj existante.
  if ( n<deg() ) 
    { 
      coeffs[n]= a; 
      return; 
    }
  
  // Troisime cas : l'indice est gal au degr, il faut distinguer :
  // - Si a est non nul, on redfinit le coeff[n] simplement.
  // - Si a est nul, le degr baisse et il faut normaliser.
  if ( n==deg() ) 
    { 
      if ( !zero(a) ) 
	{
	  coeffs[n]= a;
	  return;
	}
      else
	{ 
	  coeffs.pop_back();
	  normaliser(); 
	  return; 
	}
    }

  // Dernier cas : l'indice est plus grand que le degr :
  // - Si a est non nul il faut rallonger le vecteur.
  // - Si a est nul on laisse tomber.
  if ( zero(a) ) return;
  coeffs.resize( n+1, A(0) );
  coeffs[n]= a;
}

//-----------------------------------------------------------------------
// Addition de deux polynmes (calcul "en place")
// Entre : le polynme *this et un deuxime polynme p
// Sortie : la somme des deux, stocke dans la mmoire de *this
//-----------------------------------------------------------------------
template <typename A>
Polynome<A> Polynome<A>::operator+ () const
{ return *this; }

template <typename A>
Polynome<A>& Polynome<A>::operator+= ( const Polynome<A>& p )
{
  Degre taille= max( coeffs.size(), p.coeffs.size() );
  coeffs.resize( taille, A(0) );
  for ( Degre i=0; i<taille; ++i ) coeffs[i]+= p[i];
  normaliser();
  return *this;
}

template <typename A>
Polynome<A>& Polynome<A>::operator+= ( const A& a )
{
  if ( zero(a) ) return *this;
  if ( coeffs.empty() ) { coeffs.push_back(a); return *this; };
  coeffs[0]+= a;
  normaliser();
  return *this;
}

//-----------------------------------------------------------------------
// Addition de deux polynmes (sous forme d'oprateur binaire)
// Entre : deux polynmes p et q
// Sortie : leur somme p+q
//-----------------------------------------------------------------------
template <typename A>
Polynome<A> operator+ ( Polynome<A> p, const Polynome<A>& q )
{ return ( p += q ); }

template <typename A>
Polynome<A> operator+ ( Polynome<A> p, const A& a )
{ return ( p += a ); }

template <typename A>
Polynome<A> operator+ ( const A& a, Polynome<A> p )
{ return ( p += a ); }
  
template <typename A>
Polynome<A> operator+ ( Polynome<A> p, int a )
{ return ( p += A(a) ); }

template <typename A>
Polynome<A> operator+ ( int a, Polynome<A> p )
{ return ( p += A(a) ); }
  
//-----------------------------------------------------------------------
// Oppos d'un polynme
// Entre : le polynme *this
// Sortie : le polynme oppos
//-----------------------------------------------------------------------
template <typename A>
Polynome<A> Polynome<A>::operator- () const
{
  Polynome<A> oppose;
  Degre taille= coeffs.size();
  oppose.coeffs.resize( taille );
  for ( Degre i=0; i<taille; ++i )
    oppose.coeffs[i]= -coeffs[i];
  return oppose;
}

//-----------------------------------------------------------------------
// Soustraction de deux polynmes (calcul "en place")
// Entre : le polynme *this et un deuxime polynme p
// Sortie : la diffrence des deux, stocke dans la mmoire de *this
//-----------------------------------------------------------------------
template <typename A>
Polynome<A>& Polynome<A>::operator-= ( const Polynome<A>& p )
{
  Degre taille= max( coeffs.size(), p.coeffs.size() );
  coeffs.resize( taille, A(0) );
  for ( Degre i=0; i<taille; ++i ) coeffs[i]-= p[i];
  normaliser();
  return *this;
}

template <typename A>
Polynome<A>& Polynome<A>::operator-= ( const A& a )
{
  if ( zero(a) ) return *this;
  if ( coeffs.empty() ) { coeffs.push_back(-a); return *this; };
  coeffs[0]-= a;
  normaliser();
  return *this;
}

//-----------------------------------------------------------------------
// Soustraction de deux polynmes (sous forme d'oprateur binaire)
// Entre : deux polynmes p et q
// Sortie : leur diffrence p-q
//-----------------------------------------------------------------------
template <typename A>
Polynome<A> operator- ( Polynome<A> p, const Polynome<A>& q )
{ return ( p -= q ); }

template <typename A>
Polynome<A> operator- ( Polynome<A> p, const A& a )
{ return ( p -= a ); }

template <typename A>
Polynome<A> operator- ( const A& a, const Polynome<A>& p )
{ return ( Polynome<A>(a) -= p ); }

template <typename A>
Polynome<A> operator- ( Polynome<A> p, int a )
{ return ( p -= A(a) ); }

template <typename A>
Polynome<A> operator- ( int a, const Polynome<A>& p )
{ return ( Polynome<A>(a) -= p ); }

//-----------------------------------------------------------------------
// Multiplication de deux polynmes (par la mthode scolaire)
// Entre : deux polynmes p et q
// Sortie : leur produit p*q
//-----------------------------------------------------------------------
template <typename A>
Polynome<A> operator* ( const Polynome<A>& p, const Polynome<A>& q )
{
  if ( p.empty() || q.empty() ) return Polynome<A>();
  vector<A> coeffs( p.deg() + q.deg() + 1, A(0) );
  for ( Degre i=0; i<=p.deg(); ++i )
    for ( Degre j=0; j<=q.deg(); ++j )
      coeffs[i+j]+= p[i]*q[j];
  return Polynome<A>( coeffs );
}

//-----------------------------------------------------------------------
// Multiplication d'un polynme par une constante
// Entre : le polynme *this et une constante a
// Sortie : leur produit, stock dans la mmoire de *this
//-----------------------------------------------------------------------
template <typename A>
Polynome<A>& Polynome<A>::operator*= ( const A& a )
{
  if ( zero(a) ) { coeffs.clear(); return *this; };
  for ( Degre i=0; i<=deg(); ++i ) coeffs[i]*= a;
  return *this;
}

template <typename A>
Polynome<A> operator* ( Polynome<A> p, const A& a )
{ return ( p *= a ); }

template <typename A>
Polynome<A> operator* ( const A& a, Polynome<A> p )
{ return ( p *= a ); }

template <typename A>
Polynome<A> operator* ( Polynome<A> p, int a )
{ return ( p *= A(a) ); }

template <typename A>
Polynome<A> operator* ( int a, Polynome<A> p )
{ return ( p *= A(a) ); }

//-----------------------------------------------------------------------
// valuation d'un polynme par la mthode de Horner
// Entre : un polynme p et une constante a
// Sortie : la constante p(a) cd on remplace l'indtermine X par a
//-----------------------------------------------------------------------
template <typename A>
A Polynome<A>::operator() ( const A& a ) const
{
  A resultat( dom() );
  for ( Degre i= deg()-1; i>=0; --i )
    {
      resultat*= a; 
      resultat+= coeffs[i];
    };
  return resultat;
}

//-----------------------------------------------------------------------
// valuation d'un polynme : substitution de la variable
// Entre : un polynme q et un polynme p
// Sortie : le polynme q(p) cd on remplace l'indtermine X par p
//-----------------------------------------------------------------------
template <typename A>
Polynome<A> Polynome<A>::operator() ( const Polynome<A>& p ) const
{
  Polynome<A> resultat( dom() );
  for ( Degre i= deg()-1; i>=0; --i )
    {
      resultat*= p; 
      resultat+= coeffs[i];
    };
  return resultat;
}

//-----------------------------------------------------------------------
// Puissance d'un polynme (par la mthode dichotomique)
// Entre : un polynme p et un nombre naturel n
// Sortie : la puissance p^n
//-----------------------------------------------------------------------
template <typename A>
Polynome<A> operator^ ( Polynome<A> p, int n )
{
  Polynome<A> puissance( A(1) );
  while ( n > 0 )
    {
      while ( n%2 == 0 )
	{
	  p*= p;
	  n/= 2;
	};
      puissance*= p;
      --n;
    }
  return puissance;
}

//-----------------------------------------------------------------------
// Division euclidienne de deux polynmes 
//-----------------------------------------------------------------------
// Entre : deux polynmes a et b avec dom(b) inversible
// Sortie : deux polynmes q et r tels que a = bq+r avec deg(r)<deg(b)
//-----------------------------------------------------------------------
// Remarque :
// La paire q,r est unique. On appelle q le quotient et r le reste de 
// la division euclidienne de a par b. 
//-----------------------------------------------------------------------
template <typename A>
void eudiv( const Polynome<A>& a, const Polynome<A>& b, 
	    Polynome<A>& q, Polynome<A>& r )
{
  // attraper la seule erreur possible : un coefficient dominant non inversible
  A inverse, c= b.dom();
  if ( !inversible( c, inverse ) )
    {
      cerr << "Erreur : coefficient dominant non inversible !" << endl;
      return;
    };
  
  // effacer les anciennes valeurs de q et r
  r.clear();
  q.clear();
  
  // attraper le cas trivial o deg(a) < deg(b)
  Degre m= a.coeffs.size();
  Degre n= b.coeffs.size();
  if ( m < n ) { r= a; return; }
  
  // rserver de la place pour le quotient
  Degre j= m - n + 1;
  q.coeffs.resize( j );

  // rserver de la place et initialiser le reste
  r.coeffs.resize( n );
  for ( Degre i=0; i<n-1; ++i )
    r.coeffs[i]= a.coeffs[j+i];

  // effectuer la division euclidienne
  for( --j ; j>=0; --j )
    {
      for( Degre i=n-1; i>0; --i ) r.coeffs[i]= r.coeffs[i-1];
      r.coeffs[0]= a.coeffs[j];
      q.coeffs[j]= r.coeffs[n-1] * inverse;
      for( Degre i=n-1; i>=0; --i )
	r.coeffs[i]-= q.coeffs[j] * b.coeffs[i];
    };

  // normaliser le reste en supprimant d'ventuels zros terminaux
  r.normaliser();
  return;
}

template <typename A>
Polynome<A> eudiv( const Polynome<A>& a, const Polynome<A>& b )
{
  Polynome<A> q,r;
  eudiv(a,b,q,r);
  return q;
}

template <typename A>
Polynome<A> eumod( const Polynome<A>& a, const Polynome<A>& b )
{
  Polynome<A> q,r;
  eudiv(a,b,q,r);
  return r;
  // Ici il serait plus efficace de rimplmenter une division qui
  // ne construit que le reste r mais pas le quotient q... plus tard ?
}

//-----------------------------------------------------------------------
// Pseudo-division euclidienne de deux polynmes 
//-----------------------------------------------------------------------
// Entre : deux polynmes a et b avec b non nul.
//          On note c le coefficient dominant de b.
// Sortie : deux polynmes q et r et un entier k>=0 tels que 
//          c^ka = bq + r  avec  deg(r) < deg(b)  et  k minimal.
//-----------------------------------------------------------------------
// Remarque :
// tant donn deux polynmes a et b avec b unitaire (c=1 donc b non nul)
// il existe une unique paire de polynmes (q,r) tel que a = qb + r avec
// deg(r) < deg(b). On appelle q le quotient et r le reste de la division
// euclidienne de a par b.  Pour ce rsultat gnral il est ncessaire 
// que le coefficient dominant c de b soit inversible. Dans le cas o c 
// est non nul seulement, on obtient au moins  c^ka = bq+r avec un 
// certain exposant k>=0.  On choisira k minimal, bien sr. Pour k=0 
// on retombe sur le cas de la vraie division euclidienne.
//-----------------------------------------------------------------------
template <typename A>
int pseudo_division ( const Polynome<A>& a, const Polynome<A>& b, 
		      Polynome<A>& q, Polynome<A>& r )
{
  // attraper la seule erreur possible : la division par zro
  if ( zero(b) )
    {
      cerr << "Erreur : la division euclidienne par zro n'est pas dfinie !" << endl;
      return -1;
    };

  // effacer les anciennes valeurs de q et r
  r.clear();
  q.clear();

  // attraper le cas trivial o deg(a) < deg(b)
  Degre m= a.coeffs.size();
  Degre n= b.coeffs.size();
  if ( m < n ) { r= a; return 0; }
  
  // rserver de la place pour le quotient
  Degre j= m - n + 1;
  q.coeffs.resize( j );

  // rserver de la place et initialiser le reste
  r.coeffs.resize( n );
  for ( Degre i=0; i<n-1; ++i )
    r.coeffs[i]= a.coeffs[j+i];

  // initialiser des variables auxiliaires
  A c= b.dom();   // coefficient dominant de b
  A d;            // coefficient courant de q
  int k= 0;       // l'exposant de la pseudo-division
  A f(1);         // le facteur dj accumul c^k

  // effectuer la pseudo-division euclidienne
  for( --j ; j>=0; --j )
    {
      // dcaler le reste temporaire et ajouter le prochain coeff
      for( Degre i=n-1; i>0; --i ) r.coeffs[i]= r.coeffs[i-1];
      r.coeffs[0]= f * a.coeffs[j];

      // diviser les coefficients dominants
      if ( divisible( r.coeffs[n-1], c, d ) )
	{
	  q.coeffs[j]= d;
	  for( Degre i=n-1; i>=0; --i )
	    r.coeffs[i]-= d * b.coeffs[i];
	}
      else // attraper le cas de deux coeffs non divisibles
	{
	  ++k;
	  f*= c;
	  d= r.coeffs[n-1];
	  q.coeffs[j]= d;
	  Degre taille= q.coeffs.size();
	  for( Degre i=j+1; i<taille; ++i )
	    q.coeffs[i]*= c;
	  for( Degre i=n-1; i>=0; --i )
	    {
	      r.coeffs[i]*= c;
	      r.coeffs[i]-= d * b.coeffs[i];
	    };
	};
    };
  // normaliser le reste en supprimant les zros terminaux
  r.normaliser();
  return k;
}

template <typename A>
int pseudo_quotient( const Polynome<A>& a, const Polynome<A>& b, Polynome<A>& q )
{
  Polynome<A> r;
  return pseudo_division(a,b,q,r);
}

template <typename A>
int pseudo_reste( const Polynome<A>& a, const Polynome<A>& b, Polynome<A>& r )
{
  Polynome<A> q;
  return pseudo_division(a,b,q,r);
  // Ici il serait plus efficace de rimplmenter une pseudo-division qui
  // ne construit que le reste r mais pas le quotient q... plus tard ?
}

//-----------------------------------------------------------------------
// Fonctions de base requises pour travailler avec l'anneau Polynome<A>
//-----------------------------------------------------------------------
// Reconnatre l'lment 0
template <typename A>
bool zero( const Polynome<A>& p )
{ 
  return p.empty();
}

// Reconnatre l'lment 1
template <typename A>
bool one( const Polynome<A>& p )
{ 
  if ( p.deg() != 0 ) return false;
  return one(p[0]);
}

// Dterminer si p est inversible
template <typename A>
bool inversible( const Polynome<A>& p )
{ 
  if ( p.deg() != 0 ) return false;
  return inversible( p[0] );
}

// Dterminer si p est inversible, si oui calculer son inverse q
template <typename A>
bool inversible( const Polynome<A>& p, Polynome<A>& q )
{ 
  if ( p.deg() != 0 ) return false;
  return inversible( p[0], q );
}

// Dterminer si p est divisible par q
template <typename A>
bool divisible( const Polynome<A>& p, const Polynome<A>& q )
{
  if ( zero(p) ) return true;
  if ( zero(q) ) return false;
  Polynome<A> r;
  if ( pseudo_reste(p,q,r) > 0 ) return false;
  return zero(r);
}

// Dterminer si a est divisible par b, si oui calculer leur quotient c
template <typename A>
bool divisible( const Polynome<A>& a, const Polynome<A>& b, Polynome<A>& c )
{ 
  if ( zero(a) ) { c= a; return true; };
  if ( zero(b) ) return false;
  Polynome<A> q,r; 
  if ( pseudo_division(a,b,q,r) > 0 ) return false; 
  c= q;
  return zero(r);
}

// Dterminer si p et q sont associs
template <typename A>
bool associes( const Polynome<A>& p, const Polynome<A>& q )
{ 
  if ( p.deg() != q.deg() ) return false;
  if ( zero(p) ) return true;
  A u;
  if ( !associes( p.dom(), q.dom(), u ) ) return false;
  for ( Degre i=0; i<p.deg(); ++i ) 
    if ( u * p[i] != q[i] ) return false;
  return true;
}

// Dterminer si p et q sont associs, si oui calculer u tel que up=q
template <typename A>
bool associes( const Polynome<A>& p, const Polynome<A>& q, A& u )
{ 
  if ( p.deg() != q.deg() ) return false;
  if ( zero(p) ) { u= A(1); return true; };
  if ( !associes( p.dom(), q.dom(), u ) ) return false;
  for ( Degre i=0; i<p.deg(); ++i ) 
    if ( u * p[i] != q[i] ) return false;
  return true;
}

// Dterminer si p et q sont associs, si oui calculer u tel que up=q
template <typename A>
bool associes( const Polynome<A>& p, const Polynome<A>& q, Polynome<A>& u )
{ 
  if ( p.deg() != q.deg() ) return false;
  if ( zero(p) ) { u= A(1); return true; };
  A v;
  if ( !associes( p.dom(), q.dom(), v ) ) return false;
  for ( Degre i=0; i<p.deg(); ++i ) 
    if ( u * p[i] != q[i] ) return false;
  u= v;
  return true;
}

// Choisir un reprsentant prfr parmi les lments associs  p
template <typename A>
Polynome<A> assopref ( const Polynome<A>& p )
{
  if ( zero(p) ) return p;
  A u;
  assopref( p.dom(), u );
  return u * p;
}

// Choisir un reprsentant prfr r=up parmi les lments associs  p
template <typename A>
Polynome<A> assopref ( const Polynome<A>& p, A& u )
{
  if ( zero(p) ) { u= A(1); return p; };
  assopref( p.dom(), u );
  return u * p;
}

// Choisir un reprsentant prfr r=up parmi les lments associs  p
template <typename A>
Polynome<A> assopref ( const Polynome<A>& p, Polynome<A>& u )
{
  if ( zero(p) ) { u= A(1); return p; };
  A v;
  assopref( p.dom(), v );
  u = A(v);
  return v * p;
}

//-----------------------------------------------------------------------
// Passer d'un polynme  son successeur dans l'ordre lexicographique.
// Cela n'a qu'un sens que dans l'anneau (Z/n)[X] avec un entier n>1.
//-----------------------------------------------------------------------
template <typename A>
void avancer( Polynome<A>& p )
{
  Degre d= p.deg();
  for( Degre i=0; i<=d; ++i )
    {
      A c= p[i] + A(1);
      p.set_coeff( i, c );
      if ( !zero(c) ) return;
    }
  p.set( A(1), d+1 );
}

//-----------------------------------------------------------------------
// Calculer le contenu
// Entre : un polynme a
// Sortie : le contenu de a (le pgcd des coefficients)
//-----------------------------------------------------------------------
template <typename A>
A contenu( const Polynome<A>& a )
{
  A c= 0;
  for( Degre i=0; i<=a.deg(); ++i )
    c= pgcd(c,a[i]);
  return c;
}

//-----------------------------------------------------------------------
// Division d'un polynme par une constante
// Entre : le polynme *this et une constante a
// Sortie : leur quotient, stock dans la mmoire de *this
//-----------------------------------------------------------------------
template <typename A>
Polynome<A>& Polynome<A>::operator/= ( const A& a )
{
  A inverse;
  if ( !inversible(a,inverse) )
    cerr << "Erreur : diviseur non inversible !" << endl;
  else
    for ( Degre i=0; i<=deg(); ++i ) coeffs[i]*= inverse;
  return *this;
}

template <typename A>
Polynome<A> operator/ ( Polynome<A> a, const A& b )
{
  return ( a/= b );
}

//-----------------------------------------------------------------------
// Calculer le pgcd dans A[X] sur un anneau intgre A
// Entre : deux polynmes a et b
// Sortie : un polynme d qui est un pgcd de a et b
//-----------------------------------------------------------------------
// On suppose ici que l'anneau A est intgre et admet une fonction pgcd. 
// On l'utilise pour calculer le pgcd de deux polynmes sur A.
//-----------------------------------------------------------------------
template <typename A>
Polynome<A> pseudo_pgcd( Polynome<A> a, Polynome<A> b )
{
  // Attraper les cas triviaux
  if ( zero(a) ) return assopref(b);
  if ( zero(b) ) return assopref(a);
  
  // Dterminer les contenus et rendre a,b primitifs
  A x= contenu(a); a/= x;
  A y= contenu(b); b/= y;

  // Appliquer l'algorithme d'Euclide
  Polynome<A> c;
  while( !zero(b) )
    {
      pseudo_reste(a,b,c);
      a= b;
      b= c;
      b/= contenu(b);
    };
  return assopref( pgcd(x,y) * a );
}

//-----------------------------------------------------------------------
// Afficher un monme sous la forme cXd reprsentant cX^d. 
//-----------------------------------------------------------------------
// Option :
// Si le boolen `signe' est vrai, le monme s'affiche sous la forme cXd,
// cd on ajoute `+' si le coeff ne commence pas encore par `+' ou `-'.
//-----------------------------------------------------------------------
// Exceptions : 
// - si c == 0 on supprime l'affichage du monme
// - si d == 0 on crit c   au lieu de  cX0 (resp. +c   au lieu de +cX0)
// - si d == 1 on crit cX  au lieu de  cX1 (resp. +cX  au lieu de +cX1)
// - si c == 1 on crit  Xd au lieu de  1Xd (resp.  +Xd au lieu de +1Xd)
// - si c ==-1 on crit -Xd au lieu de -1Xd
//-----------------------------------------------------------------------
// Remarque :
// Ici la lettre `X' reprsente l'indtermine ; elle n'est utile que
// pour l'affichage, non pour la reprsentation interne. Cette lettre 
// peut donc tre choisie librement pour faciliter la lisibilit.
//-----------------------------------------------------------------------
template <typename A>
void ecrire_monome( ostream& out, bool signe, const A& coeff, char leSymbole, Degre deg )
{
  if ( zero(coeff) ) return;
  if ( one(coeff) && deg>0 )
    {
      if ( signe ) out << '+';
    }
  else if ( coeff==A(-1) && deg>0 )
    {
      out << '-';
    }
  else if ( signe )
    {
      ostringstream coeffstr;
      coeffstr << coeff << ends;
      char debut= coeffstr.str()[0];
      if ( debut != '+' && debut != '-' ) out << '+';
      out << coeffstr.str();
    }
  else out << coeff;
  if ( deg>=1 ) out << leSymbole;
  if ( deg<=1 ) return;
  out << deg;
  return;
}

//-----------------------------------------------------------------------
// Oprateur de sortie pour un polynme
//-----------------------------------------------------------------------
// Format :
// Un polynme est affich ici comme la somme de ses monmes non nuls,
// c'est--dire d'ventuels monmes nuls sont supprims. Si cette somme 
// consiste en plusieurs monmes, elle est dlimite par des parenthses.
// Une somme de longueur 1 (un seul monme) est affich sans parenthses,
// la somme vide (le polynme nul) est affich comme 0.
//-----------------------------------------------------------------------
template <typename A>
ostream& operator<< ( ostream& out, const Polynome<A>& p )
{
  // Le polynme nul est affich sans parenthses
  if ( p.deg() < 0  ) return ( out << "0" );
  
  // Chercher le premier monme non nul : est-il seul ?
  Degre i=0, deg= p.coeffs.size()-1; 
  while ( zero(p.coeffs[i]) ) ++i;
  if ( i == deg )
    {
      ecrire_monome( out, false, p[i], Polynome<A>::leSymbole, i );
      return out;
    };
  
  // Affichage la somme des monmes en supprimant les monmes nuls
  out << "(";
  ecrire_monome( out, false, p[i], Polynome<A>::leSymbole, i );
  Degre taille= p.coeffs.size();
  for ( ++i; i<taille; ++i )
    ecrire_monome( out, true, p[i], Polynome<A>::leSymbole, i );
  out << ")";
  return out;
}

//-----------------------------------------------------------------------
// Lire un monme sous la forme  cXd  reprsentant  cX^d 
// Comme toujours, l'entre doit reconnatre au moins le format produit 
// par la sortie. De plus sont accepts les formats suivants :
// X, Xd, X^d, cX^d, ainsi que X, Xd, X^d, cX^d.
//-----------------------------------------------------------------------
template <typename A>
bool lire_monome( istream& in, A& coeff, char leSymbole, Degre& deg )
{
  // remettre le flot  l'tat normal (si possible)
  in.clear();
  if ( in.fail() ) return false;
  
  // attraper un ventuel signe global
  char symbole;
  in >> symbole;
  if ( in.fail() ) return false;
  bool negatif= false;
  if ( symbole == '-' ) negatif= true;
  else if ( symbole == '+' ) negatif= false;
  else in.putback( symbole );

  // lire le prochain symbole pour distinguer entre Xd et cXd
  in >> symbole;
  if ( in.fail() ) return false;
  if ( symbole == leSymbole )
    { 
      // accepter le format Xd qui sous-entend c=1
      coeff= 1;
      in.putback( symbole );
    }
  else
    { 
      // supposer le format cXd et lire le coefficient c
      in.putback( symbole );
      in >> coeff;
      if ( in.fail() ) return false;
    };

  // tenir compte d'un ventuel signe global
  if ( negatif ) coeff= -coeff;
  
  // lire la partie Xd : tester d'abord si l'on trouve X
  in >> symbole;
  if ( symbole != leSymbole )
    {
      in.putback( symbole );
      deg= 0;
      return true;
    };
  
  // lire l'exposant s'il y en a un
  in.get( symbole );
  if ( isdigit( symbole ) )
    {
      in.putback( symbole );
      in >> deg;
      return in.good();
    };

  // accepter un ventuel symbole '^' optionnel
  if ( isspace( symbole ) )
    in >> symbole;
  if ( symbole == '^' )
    {
      in >> deg;
      return in.good();
    };

  // la lettre X n'est pas suivi d'exposant 
  in.putback( symbole );
  deg= 1;
  return true;
}

//-----------------------------------------------------------------------
// Oprateur d'entre pour un polynme
// Comme toujours, l'entre doit reconnatre au moins le format produit 
// par la sortie. De plus, un signe global est accept.
//-----------------------------------------------------------------------
template <typename A>
istream& operator>> ( istream& in, Polynome<A>& p )
{
  // remettre le flot  l'tat normal (si possible)
  in.clear();
  if ( in.fail() ) return in;

  // lire un signe global (optionnel) ou la parenthse ouvrante (optionnel)
  char symbole;
  in >> symbole;
  bool negatif= false;
  if ( symbole == '-' ) negatif= true;
  else if ( symbole == '+' ) negatif= false;
  else in.putback( symbole );
  if ( in.fail() ) return in;
  
  // lire le polynme comme somme de monmes
  p.coeffs.clear();
  A coeff;
  Degre deg;
  in >> symbole;
  if ( symbole != '(' )
    { // si l'on ne trouve pas de parenthse, lire un seul monme 
      in.putback( symbole );
      if ( lire_monome( in, coeff, Polynome<A>::leSymbole, deg ) )
	p= Polynome<A>(coeff,deg);
    }
  else
    { // sinon lire les monmes un par un, jusqu' la parenthse fermante
      for( in >> symbole; symbole!=')'; in >> symbole )
	{
	  if ( in.fail() ) return in;
	  in.putback( symbole );
	  if ( !lire_monome( in, coeff, Polynome<A>::leSymbole, deg ) )
	    {
	      in.setstate( ios::failbit ); 
	      return in;
	    };
	  if ( deg >= Degre(p.coeffs.size()) ) p.coeffs.resize( deg+1, A(0) );
	  p.coeffs[deg]+= coeff;
	};
      p.normaliser();
    };
  
  // tenir compte d'un ventuel signe global
  if ( negatif ) p= -p;
  return in;
}

#endif // __polynome_hh__

//=======================================================================
// fin du fichier polynome.hh
//=======================================================================
