//=======================================================================
// rationnel.cc : classe pour modliser le corps des nombres rationnels
// Copyright (C) 2003-2007 Michael Eisermann, all rights reserved
// Institut Fourier, Universit Joseph Fourier, Grenoble, France
//=======================================================================*/

#include <iostream>
#include "integer.cc"
using namespace std;

class Rationnel
{
private:
  Integer numer, denom;   // numrateur et dnominateur
  void normaliser(void)   // normaliser la fraction numer/denom
    { if ( zero(denom) ) { cerr << "Division par zro !\n"; exit(1); };
      Integer d= pgcd(numer,denom); numer/= d; denom/= d;
      if ( denom<0 ) { numer= -numer; denom= -denom; }; };
  
public:
  // Constructeurs divers
  Rationnel( int num=0, int den=1 )
    : numer(num), denom(den) { normaliser(); };
  Rationnel( const Integer& num, const Integer& den=1 )
    : numer(num), denom(den) { normaliser(); };
  Rationnel( const Rationnel& a )
    : numer(a.numer), denom(a.denom) {};

  // Lire le numrateur et le dnominateur
  const Integer& numerateur() const { return numer; };
  const Integer& denominateur() const { return denom; };

  // Affectation
  void affecter( const Integer& num=0, const Integer& den=1 )
    { numer= num; denom= den; normaliser(); };
  Rationnel& operator= ( const Rationnel& a )
    { numer= a.numer; denom= a.denom; return *this; };

  // Addition
  Rationnel operator+ ( const Rationnel& a ) const
    { return Rationnel( numer*a.denom + denom*a.numer, denom*a.denom ); };
  Rationnel& operator+= ( const Rationnel& a )
    { affecter( numer*a.denom + denom*a.numer, denom*a.denom ); return *this; };

  // Oppos et soustraction
  Rationnel operator- () const
    { return Rationnel( -numer, denom ); };
  Rationnel operator- ( const Rationnel& a ) const
    { return Rationnel( numer*a.denom - denom*a.numer, denom*a.denom ); };
  Rationnel& operator-= ( const Rationnel& a )
    { affecter( numer*a.denom - denom*a.numer, denom*a.denom ); return *this; };

  // Multiplication (correcte mais non optimise)
  Rationnel operator* ( const Rationnel& a ) const
    { return Rationnel( numer*a.numer, denom*a.denom ); };
  Rationnel& operator*= ( const Rationnel& a )
    { affecter( numer*a.numer, denom*a.denom ); return *this; };

  // Division (correcte mais non optimise)
  Rationnel operator/ ( const Rationnel& a ) const
    { return Rationnel( numer*a.denom, denom*a.numer ); };
  Rationnel& operator/= ( const Rationnel& a )
    { affecter( numer*a.denom, denom*a.numer ); return *this; };

  // Comparaison
  bool operator== ( const Rationnel& a ) const
    { return ( numer*a.denom == denom*a.numer ); };
  bool operator!= ( const Rationnel& a ) const
    { return ( numer*a.denom != denom*a.numer ); };
  bool operator>  ( const Rationnel& a ) const
    { return ( numer*a.denom > denom*a.numer ); };
  bool operator>= ( const Rationnel& a ) const
    { return ( numer*a.denom >= denom*a.numer ); };
  bool operator<  ( const Rationnel& a ) const
    { return ( numer*a.denom < denom*a.numer ); };
  bool operator<= ( const Rationnel& a ) const
    { return ( numer*a.denom <= denom*a.numer ); };

  // Entre et sortie
  friend istream& operator>> ( istream& in, Rationnel& a );
  friend ostream& operator<< ( ostream& out, const Rationnel& a );
};

/*-----------------------------------------------------------------------*\
 * Implmentation d'entre-sortie
\*-----------------------------------------------------------------------*/
ostream& operator<< ( ostream& out, const Rationnel& a )
{
  if ( one(a.denom) ) out << a.numer;
  else out << a.numer << "/" << a.denom;
  return out;
}

istream& operator>> ( istream& in, Rationnel& a )
{
  in.clear();
  a.affecter( Integer(0), Integer(1) );

  // Lire le numrateur
  in >> a.numer;
  if ( in.fail() ) return in;

  // Lire le symbole sparateur
  char symbole;
  in >> symbole;
  if ( in.fail() ) return in;
  if ( symbole != '/' )
    {
      in.putback(symbole);
      return in;
    };
  
  // Lire le dnominateur
  in >> a.denom;
  a.normaliser();
  return in;
}

/*-----------------------------------------------------------------------*\
 * Fonctions requises pour travailler avec l'anneau Rationnel.
 * Exercice : implmenter correctement les fonctions suivantes.
\*-----------------------------------------------------------------------*/
// Reconnatre l'lment 0
inline bool zero( const Rationnel& a )
{
  return zero(a.numerateur());
}

// Reconnatre l'lment 1
inline bool one( const Rationnel& a )
{
  return ( a.numerateur() == a.denominateur() );
}

// Dterminer si a est inversible
inline bool inversible( const Rationnel& a )
{
  return !zero(a);
}

// Dterminer si a est inversible, si oui calculer son inverse
bool inversible( const Rationnel& a, Rationnel& inverse )
{
  if ( zero(a) ) return false;
  inverse.affecter( a.denominateur(), a.numerateur() );
  return true;
}

// Dterminer si a est divisible par b
inline bool divisible( const Rationnel& a, const Rationnel& b )
{
  return zero(a) || !zero(b);
}

// Dterminer si a est divisible par b, si oui calculer le quotient q
bool divisible( const Rationnel& a, const Rationnel& b, Rationnel& q )
{
  if ( zero(a) ) { q.affecter(0); return true; };
  if ( !zero(b ) ) { q= a/b; return true; };
  return false;
}

// Dterminer si a et b sont associs
inline bool associes( const Rationnel& a, const Rationnel& b )
{
  return ( zero(a) && zero(b) ) || ( !zero(a) && !zero(b) ); 
}

// Dterminer si a et b sont associs, si oui calculer u tel que ua=b
bool associes( const Rationnel& a, const Rationnel& b, Rationnel& u )
{ 
  if ( zero(a) && zero(b) ) { u.affecter(1); return true; };
  if ( !zero(a) && !zero(b) ) { u= b/a; return true; };
  return false;
}

// Choisir un reprsentant prfr parmi les lments associs  a
inline Rationnel assopref( const Rationnel& a )
{
  if ( zero(a) ) return a; else return Rationnel(1);
}

// Choisir un reprsentant prfr r=ua parmi les lments associs  a
Rationnel assopref( const Rationnel& a, Rationnel& u )
{
  if ( zero(a) ) { u.affecter(1); return a; };
  u.affecter( a.denominateur(), a.numerateur() ); 
  return Rationnel(1);
}

// Dterminer le pgcd de a et b (trivial dans un corps)
Rationnel pgcd ( const Rationnel& a, const Rationnel& b )
{
  if ( zero(a) && zero(b) ) return Rationnel(0);
  else return Rationnel(1);
}

// Dterminer le pgcd et des coefficients u,v tels que  pgcd(a,b) = ua+vb.
Rationnel pgcd ( const Rationnel& a, const Rationnel& b, Rationnel& u, Rationnel& v )
{
  if ( !zero(a) ) { u= Rationnel(1)/a; v= Rationnel(0); return Rationnel(1); };
  if ( !zero(b) ) { u= Rationnel(0); v= Rationnel(1)/b; return Rationnel(1); };
  u= Rationnel(1); v= Rationnel(1); return Rationnel(0);
}

// Division euclidienne (tout corps est aussi un anneau euclidien)
void eudiv ( const Rationnel& a, const Rationnel& b, Rationnel& q, Rationnel& r )
{
  q= a/b; r= Rationnel(0); return;
}

// Dterminer le quotient de la division euclidienne
Rationnel eudiv ( const Rationnel& a, const Rationnel& b )
{
  return a/b;
}

// Dterminer le reste de la division euclidienne
Rationnel eumod ( const Rationnel& a, const Rationnel& b )
{
  if ( zero(b) ) return a;
  else return Rationnel(0);
}

// Dterminer la valeur absolue
Rationnel abs( const Rationnel& a ) 
{ 
  if ( a.numerateur()<0 ) return -a; else return a; 
}

/*-----------------------------------------------------------------------*\
 * Exemple d'utilisation ( mettre en commentaire aprs des tests)
\*-----------------------------------------------------------------------*/
/*
int main()
{
  cout << "Bienvenue au corps des nombres rationnels !" << endl
       << "Entrez deux lments a,b sous la forme  num/den  svp : ";
  Rationnel a,b;
  cin >> a >> b;
  cout << "a   = " << a   << endl
       << "b   = " << b   << endl
       << "a+b = " << a+b << endl
       << "a-b = " << a-b << endl
       << "a*b = " << a*b << endl
       << "a/b = " << a/b << endl;
}
*/
