//=======================================================================
// rreal-test.cc : nombres flottants fiables grces aux arrondis fiables
// Copyright (C) 2004-2007 Michael Eisermann, all rights reserved
// Institut Fourier, Universit Joseph Fourier, Grenoble, France
//
// Compilation :
// Compiler avec l'option -lgmpxx pour intgrer la bibliothque gmp. 
//
// Description :
// On teste ici des nombres flottants en prcision arbitraire.
// Ce qui nous intresse particulirement est de contrler les arrondis
// en choisissant judicieusement les modes RNDD, RNDU, RNDZ, RNDI, RNDN.
//=======================================================================

#include <iostream>
#include <iomanip>
using namespace std;

#include "rreal.cc"
using namespace Numeric;

//-----------------------------------------------------------------------
// Calculer la racine carre
// Entre : un nombre positif a 
// Sortie : un nombre positif u tel que  
//          soit  x <= u^2  avec u minimal (RNDU,RNDI)
//          soit  u^2 <= x  avec u maximal (RNDD,RNDZ)
//          soit le u le plus proche (RNDN)
//-----------------------------------------------------------------------
RReal racine( const RReal& a, Mode mode= stdMode, Len len= stdLen )
{
  // Attraper le seul cas d'erreur
  if ( a <= 0 ) return RReal(0);
  
  // On utilise d'abord le mode d'arrondi vers le haut partout.
  // On choisit une valeur initiale plus grande que la racine cherche,
  // puis on applique l'itration de Newton tant que le rsultat s'amliore.
  // On garantit ainsi une approximation du haut, de plus en plus proche.
  RReal u0, u1= (1+a)/2;
  int compteur= 0;
  do 
    { 
      // L'itration de Newton-Hron  u -> ( u + a/u ) / 2
      u1.swap( u0 );
      u1.div( a, u0, RNDU, len );
      u1.add( u0, RNDU, len );
      u1.div( 2, RNDU, len );
      
      // L'affichage ne sert que d'illustration et pourra tre supprim 
      cout << "Itration " << setw(2) << (++compteur) << " : "
	   << div( a, u1, RNDD, len ).dec( RNDD ) 
	   << " <= sqrt(x) <= " << u1.dec( RNDU ) << endl;
    } 
  while ( u1 < u0 );
  
  // On adapte la valeur finale selon le mode d'arrondi spcifi
  if( mode==RNDD || mode==RNDZ || mode==RNDN )
    u1.div( a, u1, mode, len );
  return u1;
}

//-----------------------------------------------------------------------
// La fonction main() gre la communication avec l'utilisateur
//-----------------------------------------------------------------------
int main()
{
  cout << "\nBienvenue au calcul approch de la racine carre !\n"
       << "On effectuera un calcul fiable grce aux arrondis dirigs.\n" << endl;
  Len decimales= demander_la_precision();
  cout.precision( decimales );

  /*
  // Tester la conversion de double en RReal et l'affichage
  RReal s( 0.9999999999, RNDD );
  cout << s.dec( RNDD, 3 ) << endl;
  RReal t( 0.9999999999, RNDU );
  cout << t.dec( RNDU, 3 ) << endl;
  */

  // Demander la valeur de x
  cout << "Entrer une valeur  x = ";
  RReal x;
  cin >> x;
  if ( x < 0 )
    { cout << "Au revoir.\n" << endl; return 0; };

  // Choisir l'affichage binaire pour des tests et vrifications
  // cout.unsetf( ios::dec );

  // Calcul de la racine
  cout << "\nJe calcule sqrt(x) par la mthode de Newton-Hron." << endl;
  cout << "x = " << x << endl;
  RReal mino, majo;
  majo= racine( x, RNDU );
  mino.div( x, majo, RNDD );
  
  // Affichage des rsultats avec vrification 
  cout << endl << "En choisissant judicieusement les arrondis \non obtient donc l'encadrement suivant : ";

  cout << endl << "minoration fiable = " << mino.dec( RNDD, decimales );
  mino.mul( mino, RNDU );
  if ( mino <= x ) cout << " (certifie)"; else cout << " (errone)";

  cout << endl << "majoration fiable = " << majo.dec( RNDU, decimales );
  majo.mul( majo, RNDD );
  if ( majo >= x ) cout << " (certifie)"; else cout << " (errone)";

  cout << "\n\nAu revoir et  bientt.\n" << endl;
}
