//=======================================================================
// dichotomique.cc : la mthode dichotomique en calcul arrondi fiable
// Copyright (C) 2007 Michael Eisermann, all rights reserved
// Institut Fourier, Universit Joseph Fourier, Grenoble, France
//
// Description :
// La mthode dichotomique en arithmtique d'intervalles
//
// Commentaires :
// Compiler avec l'option -lgmpxx pour intgrer la bibliothque gmp. 
//=======================================================================

#include <iostream>
#include <vector>
using namespace std;

#include "interval.cc"
using namespace Numeric;

//-----------------------------------------------------------------------
// La fonction f  analyser
//
// La difficult est d'implmentr f soigneusement pour les intervalles :
// - Il faut que le rsultat calcul f(I) contienne le vrai rsultat,
//   en tenant compte en particulier des erreurs d'arrondi.
// - Il vaut mieux que f(I) ne surencadre pas trop le vrai rsultat.
//   Sinon les rsultats sont corrects mais restent trop imprcis.
//-----------------------------------------------------------------------

//-----------------------------------------------------------------------
// Pour un exemple simple on implmente ici f(x) = x^2 - 2
//  noter que sqr(x) implmente x -> x^2 pour les intervalles.
//-----------------------------------------------------------------------
Interval f( const Interval& x, Len len= stdLen )
{
  Interval y;
  y= sqr( x, len );
  y.sub( y, 2, len );
  return y;
}

/*
//-----------------------------------------------------------------------
// Un exemple un peu plus mchant (voir le fichier chap13/bruit.cc)
// L'implmentation suivante est la traduction directe de Horner.
// Ceci est mal adapt aux intervalles et donne des surencadrements.
// Il ne faut donc pas s'tonner que la dichotomie aura du mal...
//-----------------------------------------------------------------------
Interval f( const Interval& x, Len len= stdLen )
{
  // criture concise sous forme de Horner
  // return ((((x-9)*x+30)*x-40)*x*x+48)*x-32;

  // criture plus explicite (avec le paramtre len)
  Interval y= x;
  y.add(  -9, len );  y.mul( x, len );
  y.add( +30, len );  y.mul( x, len );
  y.add( -40, len );  y.mul( x, len );
                      y.mul( x, len );
  y.add( +48, len );  y.mul( x, len );
  y.add( -32, len );
  return y;
}
*/

//-----------------------------------------------------------------------
// La mthode dichotomique en arithmtique d'intervalles
//
// Spcification :
// - On se donne un ou plusieurs intervalles initiaux  a  
//   et une marge d'erreur eps, ce qui sera la marge tolre.
// - On produit une liste de sous-intervalles de a de longueur < eps 
//   qui recouvrent toutes les racines de la fonction f dans a.
// L'essentiel ici est de garantir de ne pas louper de solutions.
// La liste finale est ensuite  analyser plus finement.  Il se peut
// que certains intervalles ne contiennent finalement pas de zro.
//
// Structure des donnes :
// On travaille avec trois listes d'intervalles :
// - La liste 'final' reoit tous les intervalles b dont on ne peut
//   exclure que f(b) contient 0 d'aprs le calcul approch, et qui
//   soient plus petits que la marge prscrite eps.
// - La liste 'nouveau' contient les intervalles encore  analyser :
//   On ne peut pas encore exclure que f(b) contient 0 et/ou 
//   la longueur est encore plus grande que eps.
// - La liste 'initiale' sert d'initialisation et stockera ensuite
//   les intervalles qui ont survecus l'itration prcdente.
//
// Mthode :
// On parcourt tous les anciens intervalles b, dont chacun est dcoup 
// en deux intervalles b1 et b2.  Si l'image f(b1) ne contient pas zro, 
// alors l'intervalle b1 est cart.  Sinon on le garde -- soit dans 
// la liste finale si b1 est assez petit, soit on le met dans la file
// d'attente 'nouveau' pour la prochaine ritration.  On traite l'autre
// l'intervalle b2 de la mme manire.
//-----------------------------------------------------------------------
void dichotomie( vector<Interval>& initial, const RReal& eps )
{
  vector<Interval> nouveau, final;
  while( !initial.empty() )
    {
      for( size_t i=0; i<initial.size(); ++i )
	{
	  const Interval& b= initial[i];
	  RReal midpoint= b.mid();
	  
	  Interval b1( b.min(), midpoint );
	  if( sign( f(b1) ) == 0 ) 
	    if( b1.len() < eps ) final.push_back(b1);
	    else nouveau.push_back(b1);
	  
	  Interval b2( midpoint, b.max() );
	  if( sign( f(b2) ) == 0 )
 	    if( b2.len() < eps ) final.push_back(b2);
	    else nouveau.push_back(b2);
	}
      swap( nouveau, initial );
      nouveau.clear();
    }
  swap( final, initial );
}

// Attendre la touche RETURN (ENTREE)
void wait_for_return()
{
  string s;
  getline(cin,s);
}

// Version verbeuse de la mthode dichotomique
void dichotomie_verbeuse( vector<Interval>& initial, const RReal& eps )
{
  cout << "Tappez ENTREE aprs chaque intration ... " << endl;
  wait_for_return();

  vector<Interval> nouveau, final;
  while( !initial.empty() )
    {
      for( size_t i=0; i<initial.size(); ++i )
	{
	  const Interval& b= initial[i];
	  cout << "Dcoupage de b=" << b << " :" << endl;
	  RReal midpoint= b.mid();

	  Interval b1( b.min(), midpoint );
	  cout << "b1=" << b1 << ", f(b1)=" << f(b1) << " ==> ";
	  if( sign( f(b1) ) == 0 ) 
	    if( b1.len() < eps ) 
	      { final.push_back(b1); cout << "final" << endl; }
	    else 
	      { nouveau.push_back(b1); cout << " itrer" << endl; }
	  else cout << "cart" << endl;

	  Interval b2( midpoint, b.max() );
	  cout << "b2=" << b2 << ", f(b2)=" << f(b2) << " ==> ";
	  if( sign( f(b2) ) == 0 ) 
	    if( b2.len() < eps ) 
	      { final.push_back(b2); cout << "final" << endl; }
	    else 
	      { nouveau.push_back(b2); cout << " itrer" << endl; }
	  else cout << "cart" << endl;
	  
	  wait_for_return();
	}
      swap( nouveau, initial );
      nouveau.clear();
    }
  swap( final, initial );
}

// La fonction main() gre la communication avec l'utilisateur
int main()
{
  cout << "\nBienvenue  l'arithmtique des intervalles !" << endl
       << "On effectue la mthode dichotomique, version fiable." << endl
       << "Entrez l'intervalle de recherche de racines svp :" << endl;
  Interval a;
  cout << "a = ";
  cin >> a;
  cout << "a = " << a << endl;
  vector<Interval> liste;
  liste.push_back( a );

  for(;;)
    {
      Len decimales= demander_la_precision();
      cout.precision( decimales );

      cout << "Entrez la marge tolre :" << endl;
      RReal eps;
      cout << "eps = ";
      cin >> eps;
      cout << "eps = " << eps << endl;
      
      dichotomie_verbeuse( liste, eps );
      cout << "\nIntervalles contenant possiblement de solutions :" << endl;
      for( size_t i=0; i<liste.size(); ++i )
	cout << liste[i] << endl;
      
      cout << "Voulez-vous raffiner ce rsultat ? (o/n) ";
      char reponse;
      cin >> reponse;
      if( reponse == 'n' ) break;
    }

  cout << "Au revoir et  bientt.\n" << endl;
}
