//=======================================================================
// gauss.cc : l'algorithme de Gauss sur les matrices
// Copyright (C) 2007 by Michael Eisermann, all rights reserved
// Institut Fourier, Universit Joseph Fourier, Grenoble, France
//
// Description : 
// On implmentera en exercice l'algorithme de Gauss et ses variantes.
//=======================================================================

#include "matrix.cc"
#include "timer.hh"
using namespace std;

//-----------------------------------------------------------------------
// limination de Gauss
//
// On transforme la matrice  a  en une matrice chelonne par des 
// oprations sur les lignes.  Cet algorithme change la matrice
// donne en paramtre.  On construit en mme temps une matrice 
// inversible  m  telle que  m*a0 = a .  Ces deux matrices sont
// rendue comme paramtres par rfrences.  La valeur renvoye 
// est le dterminant de la matrice a.
//
// Avertissement : pour le calcul numrique on applique la mthode de Gauss
// aux matrices formes de nombres  virgules flottantes.  Les erreurs 
// d'arrondi introduiront forcment des erreurs dans les matrices, puis 
// l'algorithme les propage.  Malheureusement, ce genre d'algorithme
// accumule, voire amplifie, les erreurs, de sorte qu'elles peuvent 
// exploser au cours d'un long calcul (matrices mal conditionnes).
//
// Prcaution : afin de minimiser les pertes de prcision, on essaie ici
// d'viter dj de grosses variations dans les coefficients de la matrice. 
// Tout d'abord on multiplie donc chaque ligne de sorte que le maximum 
// des valeurs absolues vaille 1 dans chaque ligne (prparation). 
// Cette astuce ne rsout pas tous les maux des erreurs d'arrondi, 
// mais augmente en gnral la prcision du rsultat. (Le tester.)
//-----------------------------------------------------------------------
template<typename T>
T gauss( Matrix<T>& a, Matrix<T>& m, bool preparation= false )
{
  //
  // Exercice (80 lignes environ)
  //
}

//-----------------------------------------------------------------------
// Tester la mthode de Gauss 
//-----------------------------------------------------------------------
template <typename T>
void test_gauss( const string fichier )
{
  // Ouvrir le fichier, afficher un message en cas d'erreur 
  ifstream in( fichier.c_str() );
  if ( !in ) 
    {
      cout << "Je ne peux pas lire le fichier \"" << fichier << "\"." << endl
	   << "Veuillez rendre la matrice accessible  cet endroit." << endl;
      return;
    }
  
  // Lire et afficher la matrice du fichier
  Matrix<T> a, u, m;
  in >> a;
  in.close();

  cout << "J'ai lu une matrice de " 
       << a.rowdim() << " lignes et "
       << a.coldim() << " colonnes : " 
       << endl << "A = " << endl << a << endl;

  cout << "\n*** Mthode de Gauss ***" << endl;
  u= a;
  Timer timer;
  timer.start();
  T det= gauss(u,m);
  timer.stop();
  cout << "temps coul : " << timer << endl;

  cout << "det(A) = " << det << endl
       << "M = " << endl << m << endl
       << "U = " << endl << u << endl
       << "M*A = " << endl << m*a << endl;
}

//-----------------------------------------------------------------------
// Un test numrique : matrices de Vandermonde
//-----------------------------------------------------------------------
template <typename T>
void test_vandermonde( void )
{
  char c;
  Matrix<T> a, u, m;
  for ( Dim n=6; n<=10; ++n )
    {
      cout << "\nMatrice de Vandermonde de dimension n = " << n << endl;
      a= vandermonde<T>(n);
      cout << "La matrice a vaut : a =" << endl << a << endl;

      cout << "On applique la mthode de Faddeev..." << endl;
      u= a;
      Timer timer;
      timer.start();
      T det= gauss(u,m);
      timer.stop();
      cout << "temps coul : " << timer << endl;

      cout << "det(A) = " << det << endl
	   << "M = " << endl << m << endl
	   << "U = " << endl << u << endl
	   << "M*A = " << endl << m*a << endl;

      cout << "Tapez ENTREE pour continuer svp : "; cin.get(c);
      system("clear");
    }
}

//-----------------------------------------------------------------------
// Un autre test numrique : matrices de Hilbert
//-----------------------------------------------------------------------
template <typename T>
void test_hilbert( void )
{
  char c;
  Matrix<T> a, u, m;
  for ( Dim n=3; n<=8; ++n )
    {
      cout << "\nMatrice de Hilbert de dimension n = " << n << endl;
      a= hilbert<T>(n);
      cout << "La matrice a vaut : a =" << endl << a << endl;

      cout << "On applique la mthode de Faddeev..." << endl;
      u= a;
      Timer timer;
      timer.start();
      T det= gauss(u,m);
      timer.stop();
      cout << "temps coul : " << timer << endl;

      cout << "det(A) = " << det << endl
	   << "M = " << endl << m << endl
	   << "U = " << endl << u << endl
	   << "M*A = " << endl << m*a << endl;

      cout << "Tapez ENTREE pour continuer svp : "; cin.get(c);
      system("clear");
    }
}

//-----------------------------------------------------------------------
// Le choix des scalaires : 
// Pour le calcul numrique on choisira en gnral le type double.
// Pour comparer avec un calcul exact on utilise ici le type Rationnel.
//  compiler avec -lgmpxx car on utilise la bibliothque GMP.
//-----------------------------------------------------------------------
#include <cmath>
typedef double Flottant;

#include <gmpxx.h>
typedef mpq_class Rationnel;

#include "rreal.cc"
using namespace Numeric;

//-----------------------------------------------------------------------
// Un exemple d'utilisation et un premier test.
//-----------------------------------------------------------------------
int main()
{
  char c; 
  system("clear");
  cout << "\nBonjour et bienvenue aux matrices et la mthode de Gauss !" << endl;

  // Exemple d'une matrice quelconque donne dans un fichier
  cout << "\n*** Mthode de Gauss utilisant des rationnels ***" << endl;
  test_gauss<Rationnel>( "exemple.mat" );
  cout << "Tapez ENTREE pour continuer svp : "; cin.get(c);
  system("clear");
  
  cout << "\n*** Mthode de Gauss utilisant des flottants ***" << endl;
  test_gauss<Flottant>( "exemple.mat" );
  cout << "Tapez ENTREE pour continuer svp : "; cin.get(c);
  system("clear");

  cout << "\n*** Mthode de Gauss utilisant la classe RReal ***" << endl;
  demander_la_precision(); cin.get(c);
  test_gauss<RReal>( "exemple.mat" );
  cout << "Tapez ENTREE pour continuer svp : "; cin.get(c);
  system("clear");

  // Exemple de matrices de Vandermonde
  cout << "\n*** Mthode de Gauss utilisant des rationnels ***" << endl;
  test_vandermonde<Rationnel>();
  cout << "\n*** Mthode de Gauss utilisant des flottants  53bits ***" << endl;
  test_vandermonde<Flottant>();
  cout << "\n*** Mthode de Gauss utilisant la classe RReal ***" << endl;
  demander_la_precision();
  test_vandermonde<RReal>();

  // Exemple de matrices de Hilbert
  cout << "\n*** Mthode de Gauss utilisant des rationnels ***" << endl;
  test_hilbert<Rationnel>();
  cout << "\n*** Mthode de Gauss utilisant des flottants  53bits ***" << endl;
  test_hilbert<Flottant>();
  cout << "\n*** Mthode de Gauss utilisant la classe RReal ***" << endl;
  demander_la_precision();
  test_hilbert<RReal>();

  cout << "Au revoir et  bientt." << endl;
}
