//=======================================================================
// graphe.cc : modlisation de graphes et de marches alatoires
// Copyright (C) 2004-2007 by Michael Eisermann, all rights reserved
// Institut Fourier, Universit Joseph Fourier, Grenoble, France
//
// En imitant l'algorithme de classement des pages web par Google,
// on implmente ici une certaine marche alatoire sur un graphe 
// de grande taille mais de faible degr.  Le classement est donn
// par l'unique mesure invariante de cette marche alatoire.
//=======================================================================

#include <iostream>    // flots d'entre-sortie (e.g. cin et cout)
#include <iomanip>     // manipulateurs pour formater les flots
#include <fstream>     // lecture et criture des fichiers
#include <vector>      // classe gnrique vector<T> 
#include <algorithm>   // algorithmes de tri etc.
#include <cmath>       // fonctions mathmatiques, e.g. abs()
#include "timer.hh"    // pour chronomtrer les calculs
using namespace std;

//-----------------------------------------------------------------------
// Implmentation de structure de graphe
//
// NB: En vue d'une modlisation d'un graphe de pages web, les sommets 
// sont aussi appels "pages" et les arrtes sont nomms "liens".
//
// Un modle mathmatique :
// Un graphe orient (fini) est une paire G = (S,A) o S est un ensemble 
// (fini) et A est une partie du produit cartsien S x S.  Les lments 
// de S sont appels "sommets" du graphe G, et un lment (j,i) de A 
// est interprt comme une "arrte" du sommet j vers le sommet i.
// 
// Un modle informatique :
// Dans l'implmentation suivante les pages sont numrots j=0,1,2,...,n-1.
// Les liens mis par la page j sont stocks comme un vecteur g[j] qui
// numre simplement les pages d'arrive i=g[j][0],g[j][1],g[j][2],...
// dans l'ordre croisant.
//-----------------------------------------------------------------------
typedef unsigned int   Page;   // les pages forment les sommets du graphe 
typedef vector<Page>   Liste;  // les liens forment les arrtes (orientes)
typedef vector<Liste>  Graph;  // les pages et les liens forment le graphe
typedef vector<double> Mesure; // mesure de probabilit sur les sommets

//-----------------------------------------------------------------------
// Oprateurs d'entre-sortie
//
// Pour l'entre-sortie les sommets sont numrots par 0,1,2,...,n-1.
// On numre les liens emis de la page j : g[j][0],g[j][1],g[j][2],...
// Ceci donne par exemple Graphe( 0:0; 1:2,3,4; 2:3,4; 3:1,4; 4:1,3; )
//-----------------------------------------------------------------------
ostream& operator<< ( ostream& out, const Graph& g )
{
  //
  // Exercice (12 lignes environ)
  //
} 

istream& operator>> ( istream& in, Graph& g )
{
  g.clear();

  // Les donnes commencent par le mot-cl `Graphe(
  string mot;
  in >> mot;
  if( mot != "Graphe(" ) 
    {
      cerr << "Erreur : absence du mot-cl `Graphe" << endl;
      return in;
    };

  // Lire les donnes jusqu' la paranthse `) terminale
  char c;
  for(;;)
    {
      // Lire d'abord le numro de la page de dpart
      Page p;
      in >> p;
      if( g.size() <= p ) g.resize( p+1 );

      // Lire le sparateur `:
      in >> c;
      if( c != ':' ) 
	cerr << "Avertissement : absence du sparateur `:" << endl;

      // Lire les pages d'arrive pour les ajouter  la liste
      for(;;)
	{
	  Page q;
	  in >> q;
	  g[p].push_back(q);
	  if( g.size() <= q ) g.resize( q+1 );
	  
	  // Lire le sparateur
	  in >> c;
	  if( c == ';' ) break;
	  if( c == ')' ) 
	    {
	      in.putback(c);
	      break;
	    }
	  if( c != ',' )
	    {
	      cerr << "Avertissement : absence du sparateur `," << endl;
	      continue;
	    }
	}

      // Tester si l'on arrive  la parenthse `) terminale
      in >> c;
      if( c == ')' ) break;
      else in.putback(c);
    }

  // Vrifier l'intgrit des donnes, les corriger si ncessaire
  for( Page p=0; p<g.size(); ++p )
    {
      Liste& liste= g[p];

      // Trier la liste des pages d'arrive
      sort( liste.begin(), liste.end() );
      
      // Supprimer des doublons
      Liste::iterator end= unique( liste.begin(), liste.end() );
      liste.erase( end, liste.end() );

      // Ajouter un lien si la liste est vide
      if ( liste.empty() )
	liste.push_back(p);
    }

  return in;
} 

//-----------------------------------------------------------------------
// Calculer la mesure invariante d'une marche alatoire sur le graphe
//
// On modlise une marche alatoire comme suit : avec probabilit c>0
// on abandonne la position actuelle et se place sur une des n sommets du 
// graphe, choisi suivant l'quiprobabilit.  Sinon, avec probabilit 1-c 
// on suit une des arrtes manant du sommet actuel. La fonction suivante
// calcule l'unique mesure invariante de cette marche alatoire.
//-----------------------------------------------------------------------
void mesure_invariante( const Graph& g, Mesure& m, const double c= 0.15 )
{
  //
  // Exercice (40 lignes environ)
  // 
  // Vous pouvez, si vous voulez, introduire des fonctions auxiliaires :
  // void propagation( const Graph& g, Mesure& x );
  // double distance( const Mesure& x, const Mesure& y );
  // 
}

//-----------------------------------------------------------------------
// Comparer deux paires (page,mesure) par la deuxime coordonne
//
// Afin de se conformer  la bibliothque STL on utilise ici une classe
// comparer  la place d'une fonction. Mais l'oprateur d'valuation est
// dfini de sorte que  comparer(p,q)  s'voque comme une fonction.
//-----------------------------------------------------------------------
typedef pair<int,double> Paire;

class comparer
{
public:
  bool operator() ( const Paire& p, const Paire& q )
    { return ( p.second > q.second ); };
}; 

//-----------------------------------------------------------------------
// Calculer puis afficher la mesure invariante
//-----------------------------------------------------------------------
void analyser( const Graph& g, const double c= 0.15 )
{
  // Calculer la mesure invariante
  cout << "Je calcule la mesure invariante pour c = " << c << endl;
  Mesure m;
  Timer timer;
  timer.start();
  mesure_invariante( g, m, c );
  timer.stop();
  cout << "temps coul : " << timer << endl;
  
  // Trier les pages page ordre d'importance
  vector<Paire> liste;
  const int n= g.size();
  for( int i=0; i<n; ++i )
    liste.push_back( Paire( i, m[i] ) );
  sort( liste.begin(), liste.end(), comparer() );

  // Afficher les pages les plus importantes
  const int p= min( 20, n );
  cout << "Voici les " << p << " pages les plus importantes : " << endl;
  cout.precision(10);
  cout.setf( ios::fixed );
  for( int i=0; i<p; ++i )
    cout << setw(6) << liste[i].first << " : " 
	 << setw(15) << liste[i].second << endl;
}

//-----------------------------------------------------------------------
// Construction d'un graphe alatoire (de taille et degr donns)
//-----------------------------------------------------------------------
void random( Graph& g, size_t taille, int degre )
{
  // Effacer puis initialiser le graphe
  g.clear();
  g.resize( taille );

  // Construire le graphe en parcourant page par page
  for ( Page p=0; p<taille; ++p )
    {
      // Toute page pointe au moins vers elle-mme
      g[p].push_back( p );

      // Ajouter des liens alatoires vers d'autres pages
      for ( int q=1; q<=degre; ++q )
	g[p].push_back( rand() % taille );

      // Trier la liste et supprimer d'ventuels doublons
      sort( g[p].begin(), g[p].end() );
      Liste::iterator end= unique( g[p].begin(), g[p].end() );
      g[p].erase( end, g[p].end() );
    }
}

//-----------------------------------------------------------------------
// Exemple d'utilisation
//-----------------------------------------------------------------------
int main()
{
  cout << "\nBonjour et bienvenue  la marche alatoire sur les graphes !\n";
  Graph g;

  ifstream in( "graphe.mat" );
  if ( in ) 
    {
      cout << "J'essaie de lire le graphe du fichier \"graphe.mat\" :" << endl;
      in >> g;
      in.close();
      if ( g.size() <= 20 ) 
	cout << g << endl;
      else
	cout << "J'ai lu un graphe d'ordre " << g.size() << endl;
    }

  if ( g.empty() )
    {
      cout << "Je crerai un graphe alatoire..." << endl
	   << "Entrez la taille et le degr svp : ";
      int taille, degre;
      cin >> taille >> degre;
  
      if ( taille*degre > 10000000 )
	{
	  cout << "Un tel graphe aurait plus de 10 millions d'arrtes.\n"
	       << "Je crains que cela soit trop... Au revoir.\n";
	  return 0;
	};
      
      if ( taille < 3 || degre < 2 )
	{
	  cout << "Ce me semble trop petit... Au revoir.\n";
	  return 0;
	};
  
      cout << "Merci. Je vais construire un tel graphe..." << endl;
      random( g, taille, degre );
      cout << "Pour documentation, j'cris ce graphe dans \"graphe.mat\"." << endl;
      ofstream out( "graphe.mat" );
      out << g;
      out.close();
    };

  // cout << "Calcul des mesures invariantes..." << endl;
  for ( double c=0.30; c>=0.0; c-=0.05 ) analyser( g, c );

  cout << "Au revoir.\n";
}
