//=======================================================================
// matrix.cc : modlisation de matrices denses de taille m*n
// Copyright (C) 2004-2007 by Michael Eisermann, all rights reserved
// Institut Fourier, Universit Joseph Fourier, Grenoble, France
//
// Description : 
// Une matrice de taille m*n est une famille de mn objets, tous d'un 
// mme type T, et indexs par (i,j) avec  1 <= i <= m  et  1 <= j <= n . 
// La classe Matrix ci-dessous implmente une matrice de taille m*n :
// la cration et redfinition de telles matrices, l'accs  l'lment 
// (i,j), puis l'entre-sortie sous forme d'une liste ou d'un tableau.
//
// Exercices :
// On implmentera en exercice les oprations lmentaires.
//=======================================================================

#include <iostream>  // entre-sortie via des flots (cin et cout)
#include <fstream>   // fonctions pour lire et crire des fichiers
#include <vector>    // classe gnrique vector<T> 
#include <string>    // chanes de caractres
#include <sstream>   // chanes comme flots
using namespace std;

//-----------------------------------------------------------------------
// On dclare un type  Dim  pour les dimension d'une matrice.
// Il servira aussi pour les indices des lignes et des colonnes.  
//-----------------------------------------------------------------------
typedef unsigned short int Dim;

//-----------------------------------------------------------------------
// Implmentation de la classe Matrix 
// On opte ici pour une classe gnrique Matrix<T> o T est le type 
// des coefficients.  On crit ainsi une seule classe qui servira 
// pour Matrix<int>, Matrix<Integer>, Matrix<double>, Matrix<RReal>
// ou toute autre type de coefficients imaginable.
//-----------------------------------------------------------------------
template <typename T>
class Matrix
{
private:
  //---------------------------------------------------------------------
  // Les donnes et quelques fonctions prives (= pas d'accs externe)
  //---------------------------------------------------------------------
  Dim rows, cols;      // nombre de lignes et de colonnes
  vector<T> coeffs;    // les coefficients de la matrice
  static const T zero; // constante statique = globale  la classe

  // Accder  l'lment index par (i,j) par rfrence sur une constante
  const T& coeff( Dim i, Dim j ) const
    { return coeffs[ (i-1)*cols + (j-1) ]; };
  
  // Accder  l'lment index par (i,j) par rfrence sur une variable
  T& coeff( Dim i, Dim j )
    { return coeffs[ (i-1)*cols + (j-1) ]; };

  // Par mesure de scurit, l'accs aux lments devrait tre restreint
  void check( Dim i, Dim j ) const
    { if ( i>=1 && i<=rows && j>=1 && j<=cols ) return; 
      cerr << "Erreur : indices hors domaine !" << endl; exit(1); };
 
public:
  //---------------------------------------------------------------------
  // Constructeurs 
  //---------------------------------------------------------------------
  // Construction d'une matrice rectangulaire nulle
  Matrix( Dim rdim, Dim cdim ) 
    : rows(rdim), cols(cdim), coeffs( rdim*cdim, zero ) {};
  
  // Construction par copie d'une autre matrice
  Matrix( const Matrix& mat ) 
    : rows(mat.rows), cols(mat.cols), coeffs( mat.coeffs ) {};

  // Construction  partir d'un vecteur
  Matrix( Dim rdim, Dim cdim, const vector<T>& vec ) 
    : rows(rdim), cols(cdim), coeffs( vec ) { coeffs.resize( rows*cols, zero ); };
  
  // Construction d'une matrice diagonale (multiple de l'identit)
  Matrix( Dim dim=0, const T& c= zero ) 
    : rows(dim), cols(dim), coeffs( dim*dim, zero )
    { for( Dim i=1; i<=dim; ++i) coeff(i,i)= c; };


  //---------------------------------------------------------------------
  // Accs externe aux donnes
  //---------------------------------------------------------------------
  // Dterminer le nombre de lignes
  Dim rowdim() const 
    { return rows; };

  // Dterminer le nombre de colonnes
  Dim coldim() const 
    { return cols; };

  // Tester si une matrice est vide (de dimension 0)
  bool empty() const 
    { return rows==0 || cols==0; };
  
  // Accs en lecture : renvoie l'lment index par (i,j) comme constante
  const T& operator() ( Dim i, Dim j ) const
    { check(i,j); return coeff(i,j); };
  
  // Accs en criture : renvoie l'lment index par (i,j) comme rfrence
  T& operator() ( Dim i, Dim j )
    { check(i,j); return coeff(i,j); };
  
  //---------------------------------------------------------------------
  // Extraction d'une sous-matrice de taille et position donnes.
  //---------------------------------------------------------------------
  void get_submatrix( Dim row_begin, Dim col_begin, Matrix& sub ) const
    { 
      // Rectifier les indices donns si ncessaire
      if ( row_begin < 1 ) row_begin= 1;
      if ( col_begin < 1 ) col_begin= 1;
      
      // Copier les coefficients
      for( int i=row_begin, si=1; i<=rows && si<=sub.rows; ++i, ++si )
	for( int j=col_begin, sj=1; j<=cols && sj<=sub.cols; ++j, ++sj )
	  sub.coeff(si,sj)= coeff(i,j);
    }

  //---------------------------------------------------------------------
  // criture d'une sous-matrice de taille et position donnes.
  //---------------------------------------------------------------------
  void set_submatrix( Dim row_begin, Dim col_begin, const Matrix& sub ) 
    { 
      // Rectifier les indices donns si ncessaire
      if ( row_begin < 1 ) row_begin= 1;
      if ( col_begin < 1 ) col_begin= 1;

      // Copier les coefficients
      for( int i=row_begin, si=1; i<=rows && si<=sub.rows; ++i, ++si )
	for( int j=col_begin, sj=1; j<=cols && sj<=sub.cols; ++j, ++sj )
	  coeff(i,j)= sub.coeff(si,sj);
    }

  //---------------------------------------------------------------------
  // change et affectation
  //---------------------------------------------------------------------
  // change efficace et sr, sans allocation de mmoire
  void swap( Matrix& mat )
    { using std::swap; swap( coeffs, mat.coeffs); 
      swap( rows, mat.rows ); swap( cols, mat.cols ); }

  // Affectation ralise par une copie et un swap
  Matrix& operator= ( const Matrix& mat )
    { Matrix copie(mat); swap(copie); return *this; };

  //---------------------------------------------------------------------
  // Rdifinitions diverses
  //---------------------------------------------------------------------
  // Redfinition comme matrice vide
  void clear()
    { rows= 0; cols= 0; coeffs.clear(); };

  // Redfinition comme matrice diagonale
  void redefine( Dim dim=0, const T& c= zero ) 
    { rows= dim; cols= dim; coeffs.clear(); 
      coeffs.resize( dim*dim, zero );
      for( Dim i=1; i<=dim; ++i) coeff(i,i)= c; };

  // Redfinition comme matrice rectangulaire
  void redefine( Dim rdim, Dim cdim ) 
  { rows= rdim; cols= cdim; coeffs.clear();
    coeffs.resize( rows*cols, zero ); };

  // Redfinition  partir d'un vecteur
  void redefine( Dim rdim, Dim cdim, const vector<T>& vec ) 
  { rows= rdim; cols= cdim; coeffs= vec;
    coeffs.resize( rows*cols, zero );  };

  //---------------------------------------------------------------------
  // Comparaison (galit/ingalit)
  //---------------------------------------------------------------------
  bool operator== ( const Matrix& mat ) const
    { return ( rows == mat.rows && cols == mat.cols && coeffs == mat.coeffs ); };
  bool operator!= ( const Matrix& mat ) const
    { return ( rows != mat.rows || cols != mat.cols || coeffs != mat.coeffs ); };
};

//-----------------------------------------------------------------------
// La constante zero. Elle est utilise comme coefficient par dfaut.
//-----------------------------------------------------------------------
template <typename T>
const T Matrix<T>::zero( T(0) );

//-----------------------------------------------------------------------
// Sortie sous forme d'un tableau, les lments tant spars par , et ;
// Pour bien formater le tableau, on crit d'abord tous les lments dans 
// des chanes de caractres.  Ayant ainsi trouv la longueur maximale
// dans chaque colonne, on peut les afficher de manire aligne.
//-----------------------------------------------------------------------
template<typename T>
ostream& operator<< ( ostream& out, const Matrix<T>& mat )
{
  Dim rows= mat.rowdim();
  Dim cols= mat.coldim();
  vector<string> str( rows * cols );
  vector<int> maxlen( cols, 0 );

  // crire d'abord tous les lments dans des chanes de caractres
  for( Dim i=1; i<=rows; ++i )
    for( Dim j=1; j<=cols; ++j )
      {
	ostringstream ostr;
	ostr << mat(i,j);
	str[ (i-1)*cols + (j-1) ]= ostr.str();
	int len= str[ (i-1)*cols + (j-1) ].size();
	if ( len > maxlen[j-1] ) maxlen[j-1]= len;
      }
  
  // Afficher le tableau en alignant les colonnes
  out.width(0); out << "( ";
  for ( Dim i=1; i<=mat.rowdim(); ++i )
    {
      for ( Dim j=1; j<=mat.coldim(); ++j )
	{
	  out.width( maxlen[j-1] );
	  out << str[ (i-1)*cols + (j-1) ];
	  if ( j<mat.coldim() ) out << ", ";
	}
      if ( i<mat.rowdim() ) out << endl << "; ";
    }
  out << ") ";
  return out;
}

//-----------------------------------------------------------------------
// Entre sous forme d'une liste, les lments tant spars par , et ;
//-----------------------------------------------------------------------
template<typename T>
istream& operator>> ( istream& in, Matrix<T>& mat )
{
  // initialiser le flot d'entre et la matrice
  in.clear();
  mat.clear();

  // trouver le symbole '(' au dbut de la liste
  char symbol;
  in >> symbol;
  if ( in.fail() ) return in;
  if ( symbol != '(' ) 
    {
      in.putback(symbol);
      in.setstate( ios::failbit );
      return in;
    };

  // lire la liste des lments
  T value;
  vector<T> list;
  Dim cols=0;
  for(;;)
    {
      // lire le prochain lment et le rajouter  la liste
      in >> value;
      if ( in.fail() ) return in;
      list.push_back( value );
      
      // lire le symbole sparateur
      in >> symbol;
      if ( in.fail() ) return in;
      
      // fin de la liste ?
      if ( symbol == ')' ) 
	{
	  if ( cols==0 ) cols= list.size();
	  break;
	};
      
      // fin de la ligne ?
      if ( symbol == ';' )
	{
	  if ( cols == 0 ) cols= list.size();
	  continue;
	};
      
      // symbole non dfini ?
      if ( symbol != ',' )
	{
	  in.putback(symbol);
	  in.setstate( ios::failbit );
	  return in;
	};
    };

  // dfinir la matrice ainsi lue
  Dim rows= list.size() / cols ;
  mat.redefine( rows, cols, list );
  return in;
}

//-----------------------------------------------------------------------
// Addition de deux matrices
//-----------------------------------------------------------------------
template<typename T>
Matrix<T> operator+ ( const Matrix<T>& a, const Matrix<T>& b )
{
  //
  // Exercice (8 lignes environ)
  //
}

//-----------------------------------------------------------------------
// Soustraction de deux matrices
//-----------------------------------------------------------------------
template<typename T>
Matrix<T> operator- ( const Matrix<T>& a, const Matrix<T>& b )
{
  //
  // Exercice (8 lignes environ)
  //
}

//-----------------------------------------------------------------------
// Multiplication par un scalaire
//-----------------------------------------------------------------------
template<typename T>
Matrix<T>& operator*= ( Matrix<T>& a, const T& c )
{
  for( Dim i=1; i<=a.rowdim(); ++i )
    for( Dim j=1; j<=a.coldim(); ++j )
      a(i,j)*= c;
  return a;
}

template<typename T>
Matrix<T> operator* ( const Matrix<T>& a, const T& c )
{
  Matrix<T> produit(a);
  for( Dim i=1; i<=a.rowdim(); ++i )
    for( Dim j=1; j<=a.coldim(); ++j )
      produit(i,j)*= c;
  return produit;
}

template<typename T>
Matrix<T> operator* ( const T& c, const Matrix<T>& a )
{
  Matrix<T> produit(a);
  for( Dim i=1; i<=a.rowdim(); ++i )
    for( Dim j=1; j<=a.coldim(); ++j )
      produit(i,j)*= c;
  return produit;
}

//-----------------------------------------------------------------------
// Division par un scalaire
//-----------------------------------------------------------------------
template<typename T>
Matrix<T>& operator/= ( Matrix<T>& a, const T& c )
{
  for( Dim i=1; i<=a.rowdim(); ++i )
    for( Dim j=1; j<=a.coldim(); ++j )
      a(i,j)/= c;
  return a;
}

template<typename T>
Matrix<T> operator/ ( const Matrix<T>& a, const T& c )
{
  Matrix<T> produit(a);
  for( Dim i=1; i<=a.rowdim(); ++i )
    for( Dim j=1; j<=a.coldim(); ++j )
      produit(i,j)/= c;
  return produit;
}

//-----------------------------------------------------------------------
// Multiplication de deux matrices
//-----------------------------------------------------------------------
template<typename T>
Matrix<T> operator* ( const Matrix<T>& a, const Matrix<T>& b )
{
  //
  // Exercice (10 lignes environ)
  //
}

//-----------------------------------------------------------------------
// Calculer la trace (mme dans le cas non carr... pourquoi pas?)
//-----------------------------------------------------------------------
template<typename T>
T trace( const Matrix<T>& a )
{
  T tr(0);
  Dim n= min( a.rowdim(), a.coldim() );
  for ( Dim i=1; i<=n; ++i ) tr+= a(i,i);
  return tr;
}

//-----------------------------------------------------------------------
// Matrices de Vandermonde (pour des tests numriques)
//-----------------------------------------------------------------------
template<typename T>
Matrix<T> vandermonde( Dim n )
{
  Matrix<T> mat(n,n);
  for ( Dim i=1; i<=n; ++i )
    {
      mat(i,1)= T(1);
      for ( Dim j=2; j<=n; ++j )
	mat(i,j)= mat(i,j-1)*T(i);
    };
  return mat;
}

//-----------------------------------------------------------------------
// matrices de Hilbert (pour des tests numriques)
//-----------------------------------------------------------------------
template<typename T>
Matrix<T> hilbert( Dim n )
{
  Matrix<T> mat(n,n);
  for ( Dim i=1; i<=n; ++i )
    for ( Dim j=1; j<=n; ++j )
      mat(i,j)= T(1)/T(i+j-1);
  return mat;
}

